"""
用于操作Gaussian09软件执行计算,生成相应输入文件,解析输出文件
"""

import numpy as np
import math
import os
import re
import shutil


class Gaussian09():
    def __init__(self):
        self.Bohr2Angs = 0.529177249
        self.pattern_energy = re.compile(r' SCF Done:  E\(.+?\)\s+=\s+(-?\d+\.\d*E?\-?\d*)\s+')
        self.pattern_energy_xtb = re.compile(r'Energy=\s*(-?\d+\.\d*E?\-?\d*)\s+')
        self.pattern_irc = re.compile(r'\s+Point Number:\s+(\d+)\s+Path Number:\s+(\d+)')

    def init_input(self, fname):
        """
        使用gjf文件初始化计算相关设置
        """
        elem, coord, headline, endline = self.parse_gjf(fname)
        self.elements = elem
        self.coord = coord
        self.head_line = headline
        self.end_line = endline

    def read_coord_from_input(self, fname):
        """
        从gjf文件中读取原子坐标
        """
        return (self.parse_gjf(fname))[1]

    def parse_gjf(self, fname):
        """
        解析gjf输入文件
        参数: 文件路径
        返回值:元素列表
               坐标(一维 numpy数组形式)
               头行
               尾行
        """
        with open(fname, 'r') as f:
            filelines = f.readlines()
        head_lines = []
        coord_lines = []
        end_lines = []
        empty_line_num = 0
        i = 0
        while i < len(filelines):
            data = filelines[i].strip().split()
            # 前两段属于关键词设置，可以不动
            if empty_line_num <= 1:
                head_lines.append(filelines[i])
            elif empty_line_num == 2:
                coord_lines.append(filelines[i])
            else:
                end_lines.append(filelines[i])
            if len(data) == 0:
                empty_line_num += 1
            i += 1
        head_lines.append(coord_lines[0])
        # 删除坐标行的电荷-多重度行和末尾的空白行
        del coord_lines[0]
        del coord_lines[-1]
        # 将尾行末尾的空行删除掉,只保留第一个空行
        end_lines_num = len(end_lines)
        for i in range(end_lines_num-1, -1, -1):
            if len(end_lines[i].strip().split()) == 0:
                del end_lines[i]
            else:
                break
        # 将元素和坐标解析出来
        element, coordinate = self.parse_coord(coord_lines)
        return (element, coordinate, head_lines, end_lines)

    def parse_coord(self, data_lines):
        """
        解析坐标和元素
        """
        elements = []
        coordinate = []
        for line in data_lines:
            data = line.strip().split()
            elements.append(data[0])
            coordinate.append([float(item) for item in data[1:]])
        return (elements, np.array(coordinate, dtype=np.float64).flatten())

    def generate_input(self, elements, coordinates, outputfilename, type='force', chk=True, oldchk=None):
        """
        写入gjf文件
        """
        coordinates = coordinates.reshape(-1, 3)
        lines = []
        for i in range(len(elements)):
            lines.append('{} {:.8f} {:.8f} {:.8f}'.format(elements[i], coordinates[i][0],
                                                          coordinates[i][1], coordinates[i][2])+'\n')
        new_head_line = [item for item in self.head_line]
        # 默认在第三行插入chk文件信息
        chkfilename = outputfilename[:-3] + 'chk'
        # 检查是否存在chk文件
        if not chk:
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    del new_head_line[i]
                    break
        else:
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    del new_head_line[i]
                    break
            new_head_line.insert(2, '%chk=' + chkfilename + '\n')
        # 写入oldchk
        if oldchk is None:
            for i in range(len(new_head_line)):
                if '%oldchk' in new_head_line[i].lower():
                    del new_head_line[i]
                    break
        else:
            for i in range(len(new_head_line)):
                if '%oldchk' in new_head_line[i].lower():
                    del new_head_line[i]
                    break
            new_head_line.insert(2, '%oldchk=' + oldchk + '\n')
        if type == 'ts':
            for i in range(len(new_head_line)):
                if 'force' in new_head_line[i].lower():
                    new_head_line[i] = new_head_line[i].lower().replace('force', 'opt=(ts,calcfc,noeigen) freq')
        elif type == 'freq':
            for i in range(len(new_head_line)):
                if 'force' in new_head_line[i].lower() and '#' in new_head_line[i].lower():
                    if oldchk is not None:
                        new_head_line[i] = new_head_line[i].lower().replace('force', 'freq guess=tcheck')
                    else:
                        new_head_line[i] = new_head_line[i].lower().replace('force', 'freq')
        elif type == 'irc':
            for i in range(len(new_head_line)):
                if 'force' in new_head_line[i].lower():
                    new_head_line[i] = new_head_line[i].lower().replace('force', 'irc=(calcfc,lqa,maxpoint=50)')
        elif type == 'force':
            if oldchk is not None:
                for i in range(len(new_head_line)):
                    if '#' in new_head_line[i].lower():
                        if 'force' in new_head_line[i].lower():
                            new_head_line[i] = new_head_line[i].lower().replace('force', 'force guess=tcheck')
                            break
        with open(outputfilename, 'w') as f:
            for line in new_head_line:
                f.writelines(line)
            for line in lines:
                f.writelines(line)
            f.writelines('\n')
            if len(self.end_line) == 0:
                f.writelines('\n')
            else:
                for line in self.end_line:
                    f.writelines(line)
                f.writelines('\n\n')

    def generate_input_modopt_ts(self, elements, coordinates, outputfilename, freezelist, chk=True):
        """
        生成限制性优化-过渡态搜索输入文件
        """
        coordinates = coordinates.reshape(-1, 3)
        lines = []
        for i in range(len(elements)):
            lines.append('{} {:.8f} {:.8f} {:.8f}'.format(elements[i], coordinates[i][0],
                                                          coordinates[i][1], coordinates[i][2])+'\n')
        new_head_line = [item for item in self.head_line]
        # 默认在第三行插入chk文件信息
        chkfilename = outputfilename[:-3] + 'chk'
        # 检查是否存在chk文件
        if not chk:
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    del new_head_line[i]
                    break
        else:
            chk_exist = False
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    chk_exist = True
                    break
            if not chk_exist:
                new_head_line.insert(2, '%chk=' + chkfilename + '\n')
        link1_head_line = ['--Link1--\n'] + new_head_line
        for i in range(len(new_head_line)):
            if 'force' in new_head_line[i].lower():
                new_head_line[i] = new_head_line[i].lower().replace('force', 'opt=(rcfc,modredundant,loose) guess=tcheck')
        for i in range(len(link1_head_line)):
            if 'force' in link1_head_line[i].lower():
                link1_head_line[i] = link1_head_line[i].lower().replace('force', 'opt=(ts,calcfc,noeigen,nofreeze) freq geom=check guess=tcheck')
        with open(outputfilename, 'w') as f:
            # 限制性优化部分
            for line in new_head_line:
                f.writelines(line)
            for line in lines:
                f.writelines(line)
            f.writelines('\n')
            # 冻结变量
            for line in freezelist:
                f.writelines(line)
            f.writelines('\n')
            # 基组部分
            if len(self.end_line) == 0:
                pass
            else:
                for line in self.end_line:
                    f.writelines(line)
                f.writelines('\n')
            # 过渡态搜索部分
            for line in link1_head_line:
                f.writelines(line)
            f.writelines('\n')
            if len(self.end_line) == 0:
                f.writelines('\n')
            else:
                for line in self.end_line:
                    f.writelines(line)
                f.writelines('\n\n')

    def generate_input_modopt_ts_xtb(self, elements, coordinates, outputfilename, freezelist, chk=True):
        """
        生成限制性优化-过渡态搜索输入文件,结合xtb使用
        """
        coordinates = coordinates.reshape(-1, 3)
        lines = []
        for i in range(len(elements)):
            lines.append('{} {:.8f} {:.8f} {:.8f}'.format(elements[i], coordinates[i][0],
                                                          coordinates[i][1], coordinates[i][2])+'\n')
        new_head_line = [item for item in self.head_line]
        # 默认在第三行插入chk文件信息
        chkfilename = outputfilename[:-3] + 'chk'
        # 检查是否存在chk文件
        if not chk:
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    del new_head_line[i]
                    break
        else:
            chk_exist = False
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    chk_exist = True
                    break
            if not chk_exist:
                new_head_line.insert(2, '%chk=' + chkfilename + '\n')
        link1_head_line = ['--Link1--\n'] + new_head_line
        for i in range(len(new_head_line)):
            if 'force' in new_head_line[i].lower():
                new_head_line[i] = new_head_line[i].lower().replace('force', 'opt=(rcfc,modredundant,loose)')
        for i in range(len(link1_head_line)):
            if 'force' in link1_head_line[i].lower():
                link1_head_line[i] = link1_head_line[i].lower().replace('force', 'opt=(ts,calcfc,noeigen,nofreeze) freq geom=check guess=tcheck')
        with open(outputfilename, 'w') as f:
            # 限制性优化部分
            for line in new_head_line:
                f.writelines(line)
            for line in lines:
                f.writelines(line)
            f.writelines('\n')
            # 冻结变量
            for line in freezelist:
                f.writelines(line)
            f.writelines('\n')
            # 基组部分
            if len(self.end_line) == 0:
                pass
            else:
                for line in self.end_line:
                    f.writelines(line)
                f.writelines('\n')
            # 过渡态搜索部分
            for line in link1_head_line:
                f.writelines(line)
            f.writelines('\n')
            if len(self.end_line) == 0:
                f.writelines('\n')
            else:
                for line in self.end_line:
                    f.writelines(line)
                f.writelines('\n\n')

    def generate_input_constrained_optimization(self, elements, coordinates, outputfilename, freezelist, chk=True):
        """
        生成限制性优化输入文件
        """
        coordinates = coordinates.reshape(-1, 3)
        lines = []
        for i in range(len(elements)):
            lines.append('{} {:.8f} {:.8f} {:.8f}'.format(elements[i], coordinates[i][0],
                                                          coordinates[i][1], coordinates[i][2])+'\n')
        new_head_line = [item for item in self.head_line]
        # 默认在第三行插入chk文件信息
        chkfilename = outputfilename[:-3] + 'chk'
        # 检查是否存在chk文件
        if not chk:
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    del new_head_line[i]
                    break
        else:
            chk_exist = False
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    chk_exist = True
                    break
            if not chk_exist:
                new_head_line.insert(2, '%chk=' + chkfilename + '\n')
        for i in range(len(new_head_line)):
            if 'force' in new_head_line[i].lower():
                new_head_line[i] = new_head_line[i].lower().replace('force', 'opt=(rcfc,modredundant,loose) guess=tcheck')
        with open(outputfilename, 'w') as f:
            # 限制性优化部分
            for line in new_head_line:
                f.writelines(line)
            for line in lines:
                f.writelines(line)
            f.writelines('\n')
            # 冻结变量
            for line in freezelist:
                f.writelines(line)
            f.writelines('\n')
            # 基组部分
            if len(self.end_line) == 0:
                pass
            else:
                for line in self.end_line:
                    f.writelines(line)
            f.writelines('\n\n')

    def generate_input_constrained_optimization_xtb(self, elements, coordinates, outputfilename, freezelist, chk=True):
        """
        生成限制性优化输入文件,结合xtb使用
        """
        coordinates = coordinates.reshape(-1, 3)
        lines = []
        for i in range(len(elements)):
            lines.append('{} {:.8f} {:.8f} {:.8f}'.format(elements[i], coordinates[i][0],
                                                          coordinates[i][1], coordinates[i][2])+'\n')
        new_head_line = [item for item in self.head_line]
        # 默认在第三行插入chk文件信息
        chkfilename = outputfilename[:-3] + 'chk'
        # 检查是否存在chk文件
        if not chk:
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    del new_head_line[i]
                    break
        else:
            chk_exist = False
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    chk_exist = True
                    break
            if not chk_exist:
                new_head_line.insert(2, '%chk=' + chkfilename + '\n')
        for i in range(len(new_head_line)):
            if 'force' in new_head_line[i].lower():
                new_head_line[i] = new_head_line[i].lower().replace('force', 'opt=(rcfc,modredundant,loose,nomicro)')
        with open(outputfilename, 'w') as f:
            # 限制性优化部分
            for line in new_head_line:
                f.writelines(line)
            for line in lines:
                f.writelines(line)
            f.writelines('\n')
            # 冻结变量
            for line in freezelist:
                f.writelines(line)
            f.writelines('\n')
            # 基组部分
            if len(self.end_line) == 0:
                pass
            else:
                for line in self.end_line:
                    f.writelines(line)
            f.writelines('\n\n')

    def generate_input_ts(self, elements, coordinates, outputfilename, chk=True, oldchk_filename=None, read_hess=False):
        """
        生成过渡态搜索输入文件
        """
        coordinates = coordinates.reshape(-1, 3)
        lines = []
        for i in range(len(elements)):
            lines.append('{} {:.8f} {:.8f} {:.8f}'.format(elements[i], coordinates[i][0],
                                                          coordinates[i][1], coordinates[i][2])+'\n')
        new_head_line = [item for item in self.head_line]
        # 默认在第三行插入chk文件信息
        chkfilename = outputfilename[:-3] + 'chk'
        # 检查是否存在chk文件
        if not chk:
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    del new_head_line[i]
                    break
        else:
            chk_exist = False
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    chk_exist = True
                    break
            if not chk_exist:
                new_head_line.insert(2, '%chk=' + chkfilename + '\n')
        # 插入oldchk
        if oldchk_filename is not None:
            new_head_line.insert(2, '%oldchk=' + oldchk_filename[:-3] + 'chk' + '\n')
        for i in range(len(new_head_line)):
            if 'force' in new_head_line[i].lower():
                if oldchk_filename is not None:
                    if read_hess:
                        new_head_line[i] = new_head_line[i].lower().replace('force', 'opt=(rcfc,noeigen,ts) freq guess=tcheck')
                    else:
                        new_head_line[i] = new_head_line[i].lower().replace('force', 'opt=(calcfc,noeigen,ts) freq guess=tcheck')
                else:
                    # 没有oldchk文件时无法读取Hessian
                    new_head_line[i] = new_head_line[i].lower().replace('force', 'opt=(calcfc,noeigen,ts) freq')
        with open(outputfilename, 'w') as f:
            # 文件头
            for line in new_head_line:
                f.writelines(line)
            # 坐标
            for line in lines:
                f.writelines(line)
            f.writelines('\n')
            # 基组部分
            if len(self.end_line) == 0:
                pass
            else:
                for line in self.end_line:
                    f.writelines(line)
            f.writelines('\n\n')

    def generate_input_ts_xtb(self, elements, coordinates, outputfilename, chk=True, oldchk_filename=None, read_hess=False):
        """
        生成过渡态搜索输入文件,只进行过渡态搜索,不做频率分析
        """
        coordinates = coordinates.reshape(-1, 3)
        lines = []
        for i in range(len(elements)):
            lines.append('{} {:.8f} {:.8f} {:.8f}'.format(elements[i], coordinates[i][0],
                                                          coordinates[i][1], coordinates[i][2])+'\n')
        new_head_line = [item for item in self.head_line]
        # 默认在第三行插入chk文件信息
        chkfilename = outputfilename[:-3] + 'chk'
        # 检查是否存在chk文件
        if not chk:
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    del new_head_line[i]
                    break
        else:
            chk_exist = False
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    chk_exist = True
                    break
            if not chk_exist:
                new_head_line.insert(2, '%chk=' + chkfilename + '\n')
        # 插入oldchk
        if oldchk_filename is not None:
            new_head_line.insert(2, '%oldchk=' + oldchk_filename[:-3] + 'chk' + '\n')
        for i in range(len(new_head_line)):
            if 'force' in new_head_line[i].lower():
                if oldchk_filename is not None:
                    if read_hess:
                        new_head_line[i] = new_head_line[i].lower().replace('force', 'opt=(rcfc,noeigen,ts,nomicro)')
                    else:
                        new_head_line[i] = new_head_line[i].lower().replace('force', 'opt=(calcfc,noeigen,ts,nomicro)')
                else:
                    # 没有oldchk文件时无法读取Hessian
                    new_head_line[i] = new_head_line[i].lower().replace('force', 'opt=(calcfc,noeigen,ts,nomicro)')
        with open(outputfilename, 'w') as f:
            # 文件头
            for line in new_head_line:
                f.writelines(line)
            # 坐标
            for line in lines:
                f.writelines(line)
            f.writelines('\n')
            # 基组部分
            if len(self.end_line) == 0:
                pass
            else:
                for line in self.end_line:
                    f.writelines(line)
            f.writelines('\n\n')

    def generate_input_freq_xtb(self, elements, coordinates, outputfilename, chk=True, oldchk_filename=None, read_hess=False):
        """
        生成频率分析输入文件
        """
        coordinates = coordinates.reshape(-1, 3)
        lines = []
        for i in range(len(elements)):
            lines.append('{} {:.8f} {:.8f} {:.8f}'.format(elements[i], coordinates[i][0],
                                                          coordinates[i][1], coordinates[i][2])+'\n')
        new_head_line = [item for item in self.head_line]
        # 默认在第三行插入chk文件信息
        chkfilename = outputfilename[:-3] + 'chk'
        # 检查是否存在chk文件
        if not chk:
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    del new_head_line[i]
                    break
        else:
            chk_exist = False
            for i in range(len(new_head_line)):
                if '%chk' in new_head_line[i].lower():
                    chk_exist = True
                    break
            if not chk_exist:
                new_head_line.insert(2, '%chk=' + chkfilename + '\n')
        # 插入oldchk
        if oldchk_filename is not None:
            new_head_line.insert(2, '%oldchk=' + oldchk_filename[:-3] + 'chk' + '\n')
        for i in range(len(new_head_line)):
            if 'force' in new_head_line[i].lower():
                new_head_line[i] = new_head_line[i].lower().replace('force', 'freq')
        with open(outputfilename, 'w') as f:
            # 文件头
            for line in new_head_line:
                f.writelines(line)
            # 坐标
            for line in lines:
                f.writelines(line)
            f.writelines('\n')
            # 基组部分
            if len(self.end_line) == 0:
                pass
            else:
                for line in self.end_line:
                    f.writelines(line)
            f.writelines('\n\n')

    def formchk(self, chkfilename):
        """
        将chk文件转化为fchk文件
        """
        fchk_filename = chkfilename.split('.')[0] + '.fchk'
        os.system('formchk ' + chkfilename + ' ' + fchk_filename + ' > /dev/null')

    def unformchk(self, fchkfilename):
        """
        将fchk文件转化为chk文件
        """
        chk_filename = fchkfilename.split('.')[0] + '.chk'
        os.system('unfchk ' + fchkfilename + ' ' + chk_filename + ' > /dev/null')

    def write_hess_fchk(self, hess, fchkfilename):
        """
        将笛卡尔坐标下的Hessian写入fchk文件中
        """
        n, _ = hess.shape
        hess_num = int(n * (n+1) / 2)
        head_line = 'Cartesian Force Constants                  R   N={:>12}'.format(hess_num)
        # 写入Hessian
        hess_list = []
        for line_num in range(n):
            for i in range(line_num+1):
                hess_list.append(hess[line_num, i])
        # 以每5个值为一行构造输出行
        hess_str_list = []
        for i in range(len(hess_list)):
            if i % 5 != 4 and (i != len(hess_list)-1):
                # 添加数字和换行符
                hess_str_list.append('{:>16.8e}'.format(hess_list[i]))
            else:
                hess_str_list.append('{:>16.8e}\n'.format(hess_list[i]))
        total_str = ''.join(hess_str_list)
        with open(fchkfilename, 'a') as f:
            f.writelines(head_line + '\n')
            f.write(total_str)

    def run(self, fname):
        """
        运行g09计算gjf文件
        """
        input_name = fname[:-3] + 'gjf'
        output_name = fname[:-3] + 'log'
        chk_name = fname[:-3] + 'chk'
        os.system('g09 ' + input_name + ' ' + output_name)
        # 运行完毕后检查chk文件是否存在
        if os.path.exists(chk_name):
            self.formchk(chk_name)

    def get_force(self, fname):
        """
        从输出文件中获取受力
        参数:文件名
        输出:受力(一维 numpy.array数组 单位为Hartree/Angstrom)
        """
        pattern_force_head = re.compile(r'Axes restored to original set')
        pattern_force_head = re.compile(r'Center\s+Atomic\s+Forces\s+\(Hartrees/Bohr\)')
        pattern_force_data = re.compile(r'\s+(\d+?)\s+(\d+?)\s+(-?\d+?\.\d+)\s+(-?\d+?\.\d+)\s+(-?\d+?\.\d+)')
        file_lines = []
        with open(fname, 'r') as f:
            file_lines = f.readlines()
        for i in range(len(file_lines)):
            force_head = re.search(pattern_force_head, file_lines[i])
            if force_head:
                i += 3
                # 开始搜索受力
                force_data = re.search(pattern_force_data, file_lines[i])
                force_tmp = []
                while force_data:
                    force_tmp.append([float(force_data.group(3)),
                                      float(force_data.group(4)),
                                      float(force_data.group(5))])
                    i += 1
                    force_data = re.search(pattern_force_data, file_lines[i])
        return np.array(force_tmp, np.float64).flatten()

    def get_energy(self, fname):
        """
        从输出文件中获取能量
        参数:文件名
        输出:能量(浮点数 Hartree)
        """
        lines = []
        with open(fname, 'r') as f:
            lines = f.readlines()
        energy = None
        file_length = len(lines)
        i = 0
        while i < file_length:
            result = re.search(self.pattern_energy, lines[i])
            if result:
                energy = float(result.group(1))
            i += 1
        i = 0
        if energy is None:
            while i < file_length:
                result = re.search(self.pattern_energy_xtb, lines[i])
                if result:
                    energy = float(result.group(1))
                i += 1
        return energy

    def get_coord_fchk(self, fchk_name):
        """
        从fchk文件中读取坐标
        原子单位
        """
        with open(fchk_name, 'r') as f:
            filelines = f.readlines()
        coord_data = []
        for i, line in enumerate(filelines):
            if 'Current cartesian coordinates' in line:
                data_num = int(line.strip().split()[-1])
                line_num = math.ceil(data_num / 5)
                coord_data = filelines[i+1:i+line_num+1]
                break
        coord = []
        for data_str in coord_data:
            coord.extend([float(item) for item in data_str.strip().split()])
        return np.array(coord, np.float64)

    def get_force_fchk(self, fchk_name):
        """
        从fchk文件中读取force
        原子单位
        """
        with open(fchk_name, 'r') as f:
            filelines = f.readlines()
        # 搜索Cartesian Gradient行
        grad_data = []
        for i, line in enumerate(filelines):
            if 'Cartesian Gradient' in line:
                data_num = int(line.strip().split()[-1])
                line_num = math.ceil(data_num / 5)
                grad_data = filelines[i+1:i+line_num+1]
                break
        gradient = []
        for data_str in grad_data:
            gradient.extend([float(item) for item in data_str.strip().split()])
        return -np.array(gradient, np.float64)

    def get_hess_fchk(self, fchk_name):
        """
        从fchk文件中读取Hessian矩阵
        原子单位
        """
        with open(fchk_name, 'r') as f:
            filelines = f.readlines()
        # 搜索Cartesian Gradient行
        hess_data = []
        for i, line in enumerate(filelines):
            if 'Cartesian Force Constants' in line:
                data_num = int(line.strip().split()[-1])
                line_num = math.ceil(data_num / 5)
                hess_data = filelines[i+1:i+line_num+1]
                break
        hess = []
        for data_str in hess_data:
            hess.extend([float(item) for item in data_str.strip().split()])
        hess_dim = int((np.sqrt(8*data_num+1) - 1) / 2)
        # 将下三角形式转为完整二维形式
        hessian = np.zeros((hess_dim, hess_dim), np.float64)
        for i in range(hess_dim):
            hessian[i, :i+1] = hess[int(i*(i+1)/2):int((i+1)*(i+2)/2)]
        hessian = hessian + np.tril(hessian, -1).T
        return hessian

    def check_termnation(self, fname):
        """
        检查结束状态
        """
        with open(fname, 'r') as f:
            endline = f.readlines()[-1]
        if 'Normal' in endline:
            return True
        else:
            return False

    def read_file_log_opt(self, fname):
        """
        读取优化文件并解析
        """
        pattern_coor_head = re.compile(r'\s+Input orientation:')
        pattern_coor_data = re.compile(r'\s+(\d+)\s+(\d+)\s+(\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)')
        if os.path.exists(fname):
            with open(fname, 'r') as f:
                file_lines = f.readlines()
        else:
            print('文件不存在! :', fname)
            return
        # search SPE单点能
        # for i in range(len(file_lines)):
        i = 0
        while i < len(file_lines):
            # 搜索坐标头行
            coor_head = re.search(pattern_coor_head, file_lines[i])
            if coor_head:
                i += 5
                # 开始搜索坐标
                coor_data = re.search(pattern_coor_data, file_lines[i])
                coor_tmp = []
                while coor_data:
                    coor_tmp.append([float(coor_data.group(4)),
                                    float(coor_data.group(5)),
                                    float(coor_data.group(6))])
                    i += 1
                    coor_data = re.search(pattern_coor_data, file_lines[i])
            i += 1
        return np.array(coor_tmp, 'f').flatten()

    def read_file_log_opt_all(self, fname):
        """
        读取优化文件并解析所有坐标
        """
        pattern_coor_head = re.compile(r'\s+Input orientation:')
        pattern_coor_data = re.compile(r'\s+(\d+)\s+(\d+)\s+(\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)')
        if os.path.exists(fname):
            with open(fname, 'r') as f:
                file_lines = f.readlines()
        else:
            print('文件不存在! :', fname)
            return
        # search SPE单点能
        # for i in range(len(file_lines)):
        i = 0
        all_coord = []
        while i < len(file_lines):
            # 搜索坐标头行
            coor_head = re.search(pattern_coor_head, file_lines[i])
            if coor_head:
                i += 5
                # 开始搜索坐标
                coor_data = re.search(pattern_coor_data, file_lines[i])
                coor_tmp = []
                while coor_data:
                    coor_tmp.append([float(coor_data.group(4)),
                                    float(coor_data.group(5)),
                                    float(coor_data.group(6))])
                    i += 1
                    coor_data = re.search(pattern_coor_data, file_lines[i])
                all_coord.append(np.array(coor_tmp, 'f').flatten())
            i += 1
        return all_coord

    def get_energy_all(self, fname):
        """
        获取log文件中的能量
        """
        with open(fname, 'r') as f:
            filetext = f.read()
        energy_list = []
        result = re.findall(self.pattern_energy, filetext)
        if result:
            for item in result:
                energy_list.append(float(item[1]))
        return energy_list

    def get_energy_irc(self, fname):
        """
        获取irc计算文件中的能量
        """
        energy_list = self.get_energy_all(fname)
        irc_path_A = []
        irc_path_B = []
        with open(fname, 'r') as f:
            filetext = f.read()
        result = re.findall(self.pattern_irc, filetext)
        if result:
            for i, item in enumerate(result):
                path_num = int(item[1])
                if path_num == 1:
                    irc_path_A.append(energy_list[i])
                else:
                    irc_path_B.append(energy_list[i])
        irc_path_B.reverse()
        irc_path = irc_path_B + irc_path_A
        return irc_path

    def read_file_log_irc(self, fname):
        """
        读取IRC文件并解析
        """
        pattern_coor_head = re.compile(r'\s+Input orientation:(.*\n){5}((.*\n)*?)\s*------*')
        if os.path.exists(fname):
            with open(fname, 'r') as f:
                filetext = f.read()
        else:
            print('文件不存在! :', fname)
            return
        coord_string = re.findall(pattern_coor_head, filetext)
        coord_list = []
        if coord_string:
            for coord_string_group in coord_string:
                tmp_coord = []
                data = coord_string_group[1].strip().split('\n')
                for tmp_data in data:
                    tmp_coord.append([float(item) for item in tmp_data.split()[3:]])
                coord_list.append(np.array(tmp_coord, np.float64))
        # 最后一个坐标时重复的
        return coord_list[:-1]

    def get_coord_irc(self, fname):
        """
        按顺序获取IRC文件中的坐标
        """
        coord_list = self.read_file_log_irc(fname)
        irc_path_A = []
        irc_path_B = []
        with open(fname, 'r') as f:
            filetext = f.read()
        result = re.findall(self.pattern_irc, filetext)
        if result:
            for i, item in enumerate(result):
                path_num = int(item[1])
                if path_num == 1:
                    irc_path_A.append(coord_list[i])
                else:
                    irc_path_B.append(coord_list[i])
        irc_path_B.reverse()
        irc_path = irc_path_B + irc_path_A
        return irc_path

    def trans_freq2ts(self, freq_fname, ts_fname):
        """
        将频率分析任务转化为过渡态搜索任务
        """
        old_gjf_fname = freq_fname[:-3] + 'gjf'
        new_gjf_fname = ts_fname[:-3] + 'gjf'
        old_chk_fname = freq_fname[:-3] + 'chk'
        new_chk_fname = ts_fname[:-3] + 'chk'
        # shutil.copy(old_chk_fname, new_chk_fname)
        # 将ts文件中的chk路径修改为当前ts的chk路径
        with open(old_gjf_fname, 'r') as f:
            new_txt = f.readlines()
        head_lines = []
        data_lines = []
        for i in range(len(new_txt)):
            data_str = new_txt[i].strip().split()
            if len(data_str) == 0:
                data_lines = new_txt[i:]
                break
            else:
                head_lines.append(new_txt[i])
        # 先检查chk路径
        for i in range(len(head_lines)):
            if '%chk' in head_lines[i].lower():
                # 存在chk路径,进行修改
                head_lines[i] = '%chk=' + new_chk_fname + '\n'
                continue
            if '%oldchk' in head_lines[i].lower():
                # 存在oldchk路径,进行修改
                head_lines[i] = '%oldchk=' + old_chk_fname + '\n'
                continue
            if '#' in head_lines[i] and 'freq' in head_lines[i].lower():
                # 修改freq关键词为ts关键词
                # head_lines[i] = head_lines[i].lower().replace('freq', 'opt=(rcfc,ts,noeigen,cartesian) freq guess=tcheck')
                head_lines[i] = head_lines[i].lower().replace('freq', 'opt=(rcfc,ts,noeigen) freq guess=tcheck')
                continue
        # 拼接写入
        with open(new_gjf_fname, 'w') as f:
            for line in head_lines:
                f.writelines(line)
            for line in data_lines:
                f.writelines(line)

    def trans_force2ts(self, force_fname, ts_fname):
        """
        将受力任务转化为过渡态搜索任务
        """
        old_gjf_fname = force_fname[:-3] + 'gjf'
        new_gjf_fname = ts_fname[:-3] + 'gjf'
        old_chk_fname = force_fname[:-3] + 'chk'
        new_chk_fname = ts_fname[:-3] + 'chk'
        shutil.copy(old_chk_fname, new_chk_fname)
        # 将ts文件中的chk路径修改为当前ts的chk路径
        with open(old_gjf_fname, 'r') as f:
            new_txt = f.readlines()
        head_lines = []
        data_lines = []
        for i in range(len(new_txt)):
            data_str = new_txt[i].strip().split()
            if len(data_str) == 0:
                data_lines = new_txt[i:]
                break
            else:
                head_lines.append(new_txt[i])
        # 先检查chk路径
        for i in range(len(head_lines)):
            if '%chk' in head_lines[i].lower():
                # 存在chk路径,进行修改
                head_lines[i] = '%chk=' + new_chk_fname + '\n'
            if 'force' in head_lines[i].lower():
                # 修改freq关键词为ts关键词
                head_lines[i] = head_lines[i].lower().replace('force', 'opt=(rcfc,ts,noeigen,cartesian) freq guess=tcheck')
        # 拼接写入
        with open(new_gjf_fname, 'w') as f:
            for line in head_lines:
                f.writelines(line)
            for line in data_lines:
                f.writelines(line)

    def conunt_step_opt(self, opt_filename):
        """
        统计优化任务中执行了多少次优化
        通过统计Converged出现次数进行判断
        """
        with open(opt_filename, 'r') as f:
            filetxt = f.read()
        count_str = 'Converged'
        return filetxt.count(count_str)


if __name__ == "__main__":
    path = r'D:\gaussian\Test\C2H6_opt.gjf'
    path = r'D:\gaussian\baker_reaction\DFT\EIP\irc\irc_1.log'
    fchk_path = r'D:\gaussian\Test\freq_H_0.fchk'
    g09 = Gaussian09()
    # g09.parse_gjf(path)
    # g09.get_energy_irc(path)
    # g09.get_coord_irc(path)
    hess = np.array([[1, 2, 4, 5, 6],
                     [3, 4, 2, 3, 4],
                     [3, 4, 2, 3, 4],
                     [1, 1, 1, 1, 1],
                     [1, 2, 3, 4, 5]], 'f')
    g09.write_hess_fchk(hess, fchk_path)
