"""
IDPP插值

"""
import numpy as np
import os
import copy


class IDPP():
    def __init__(self):
        self.image_list = []
        self.coord_list = []
        self.init_mat = None
        self.final_mat = None
        self.init_coord = None
        self.final_coord = None
        self.n = 0
        self.bfgs = BFGS()
        # 收敛限
        self.convergence = 1e-4

    def init_image(self, init_coord, final_coord, n):
        """
        生成像点坐标
        """
        self.n = n
        init_dis_mat = self.cal_dis_mat(init_coord)
        final_dis_mat = self.cal_dis_mat(final_coord)
        self.image_list = [init_dis_mat]
        self.init_mat = init_dis_mat
        self.init_coord = init_coord
        self.coord_list = [init_coord]
        self.final_mat = final_dis_mat
        self.final_coord = final_coord
        for i in range(n):
            image = init_dis_mat + (i+1)*(final_dis_mat-init_dis_mat)/(n+1)
            coord = init_coord + (i+1)*(final_coord-init_coord)/(n+1)
            # 坐标线性插值
            self.image_list.append(image)
            self.coord_list.append(coord)
        self.image_list.append(final_dis_mat)
        self.coord_list.append(final_coord)

    def interpolation_liner(self, init_coord, final_coord, n):
        """
        线性插值
        """
        inter_coord_list = []
        for i in range(n):
            coord = init_coord + (i+1)*(final_coord-init_coord)/(n+1)
            # 坐标线性插值
            inter_coord_list.append(coord)
        return inter_coord_list

    def cal_dis_mat(self, atom_coord):
        """
        计算距离矩阵
        参数:
            atom_coord: 原子坐标，格式: np.array([[x, y, z],[x, y, z],...], 'f')
        """
        atom_num = len(atom_coord)
        dis_mat = np.zeros((atom_num, atom_num), np.float64)
        for i in range(atom_num):
            for j in range(i, atom_num):
                if i == j:
                    dis_mat[i, j] = 1
                else:
                    dis = np.linalg.norm(atom_coord[i]-atom_coord[j])
                    # 距离矩阵是对角矩阵，对角线值为0
                    dis_mat[i, j] = dis
                    dis_mat[j, i] = dis
        return dis_mat

    def mat2vec(self, mat):
        """
        取矩阵上三角部分转化为向量
        """
        length = mat.shape[0]
        target = np.zeros(int(length*(length-1)/2), np.float64)
        start = 0
        end = 0
        for i in range(length):
            end += length-i-1
            target[start:end] = mat[i, i+1:]
            start = end
        return target

    def idpp_energy_new(self, current_coord, image_matrix):
        # 计算当前距离矩阵
        current_matrix = self.cal_dis_mat(current_coord)
        energy = np.sum((image_matrix - current_matrix)**2 / image_matrix**4) / 2
        return energy

    def idpp_energy(self, current_coord, image_matrix):
        # 计算当前距离矩阵
        current_matrix = self.cal_dis_mat(current_coord)
        energy = np.sum((image_matrix - current_matrix)**2 / current_matrix**4) / 2
        return energy

    def idpp_force_new(self, current_coord, image_matrix):
        atom_num = len(current_coord)
        force = np.zeros((atom_num, 3), np.float64)
        dist_im = image_matrix
        dist_tmp = self.cal_dis_mat(current_coord)
        coord_diff = []
        for coord in current_coord:
            diff = current_coord - coord
            coord_diff.append(diff)
        coord_diff = np.array(coord_diff, np.float64)
        grad = np.zeros((atom_num, 3), np.float64)
        # mat = (dist_im-dist_tmp)*(2*dist_im-dist_tmp)/dist_tmp**5
        dist_diff = dist_tmp - dist_im
        mat = dist_diff*(1 - 2*dist_diff/dist_tmp)/dist_im**5
        grad = 2 * (mat[..., np.newaxis] * coord_diff).sum(0)
        force = -grad.reshape(-1, 3)
        return force

    def idpp_force(self, current_coord, image_matrix):
        atom_num = len(current_coord)
        force = np.zeros((atom_num, 3), np.float64)
        dist_im = image_matrix
        dist_tmp = self.cal_dis_mat(current_coord)
        coord_diff = []
        for coord in current_coord:
            diff = current_coord - coord
            coord_diff.append(diff)
        coord_diff = np.array(coord_diff, np.float64)
        grad = np.zeros((atom_num, 3), np.float64)
        # mat = (dist_im-dist_tmp)*(2*dist_im-dist_tmp)/dist_tmp**5
        dist_diff = dist_tmp - dist_im
        mat = dist_diff*(1 - 2*dist_diff/dist_tmp)/dist_tmp**5
        grad = 2 * (mat[..., np.newaxis] * coord_diff).sum(0)
        force = -grad.reshape(-1, 3)
        return force

    def opt(self, n):
        """
        梯度下降法优化
        """
        max_step_size = 0.1
        alpha = 0.2
        for i in range(10000):
            force = self.idpp_force(self.coord_list[n], self.image_list[n])
            if np.linalg.norm(force) < 0.001:
                break
            step = force * alpha
            step_max = np.abs(step).max()
            if step_max >= max_step_size:
                step *= max_step_size / step_max
            self.coord_list[n] = self.coord_list[n] + step

    def opt_bfgs(self, n):
        """
        bfgs方法优化
        """
        self.bfgs.reset_bfgs()
        for i in range(1000):
            force = self.idpp_force(self.coord_list[n], self.image_list[n])
            if np.abs(force).max() < self.convergence:
                # print(n, i, np.abs(force).max())
                break
            self.coord_list[n][:] = self.bfgs.get_next_coord(self.coord_list[n].flatten(), force.flatten()).reshape(-1, 3)

    def align(self, i):
        """
        对齐原子坐标
        """
        if i == 0:
            center = np.sum(self.coord_list[i], axis=0) / self.coord_list[i].shape[0]
            self.coord_list[i] = self.coord_list[i] - center
            return np.eye(3, dtype=np.float64)
        else:
            target_coord = self.coord_list[i-1]
            new_coord = self.coord_list[i]
            _, new_coord, r_mat = atom_align(target_coord, new_coord)
            self.coord_list[i] = new_coord
            return r_mat

    def add_random_offset(self, i):
        """
        在垂直于切线的方向上添加随机平移
        """
        # random_offset = np.random.random(self.coord_list[0].shape)
        # random_offset = random_offset / np.linalg.norm(random_offset) * 0.1
        # 构建三个offset
        x_offset = np.zeros(self.coord_list[0].shape, np.float64).reshape(-1, 3)
        y_offset = np.zeros(self.coord_list[0].shape, np.float64).reshape(-1, 3)
        z_offset = np.zeros(self.coord_list[0].shape, np.float64).reshape(-1, 3)
        for j in range(len(x_offset)):
            x_offset[j][j%3] = 1
            y_offset[j][(j+1)%3] = 1
            z_offset[j][(j+2)%3] = 1
        x_offset = x_offset / np.linalg.norm(x_offset)
        y_offset = y_offset / np.linalg.norm(y_offset)
        z_offset = z_offset / np.linalg.norm(z_offset)
        tang_direction = self.coord_list[i] - self.coord_list[i-1]
        tang_direction = tang_direction / np.linalg.norm(tang_direction)
        # 选择向量在切线方向投影最小的那个，表明垂直度最高
        project_list = [np.abs(np.dot(item.flatten(), tang_direction.flatten())) for item in [x_offset, y_offset, z_offset]]
        index = project_list.index(min(project_list))
        offset_list = [x_offset, y_offset, z_offset]
        offset = offset_list[index].flatten()
        random_offset = offset
        random_offset = random_offset / np.linalg.norm(random_offset) * 0.1
        self.coord_list[i][:] = self.coord_list[i] + random_offset.reshape(-1, 3)


class BFGS:
    def __init__(self):
        self.alpha = 0.1
        self.converge_max_force = 4.5e-4
        self.converge_rmsd_force = 3.0e-4
        self.converge_max_displace = 1.8e-3
        self.converge_rmsd_displace = 1.2e-3
        self.force = []
        self.step = []
        self.H = None
        self.step_num = 0
        self.MAX_STEP = 0.15

    def reset_bfgs(self):
        self.force = []
        self.step = []
        self.H = None
        self.step_num = 0

    def scale_by_max_step(self, step, max_step):
        """
        对过大的步长进行修正
        """
        steps_max = np.abs(step).max()
        if steps_max > max_step:
            step = step * max_step / steps_max
        return step

    def update_hess_BFGS(self, s, y):
        """
        s: 坐标差
        y: 梯度差
        """
        if np.abs(s.dot(y)) < 1e-6:
            return
        rho = 1 / s.dot(y)
        if rho < 0:
            # self.H = np.eye(len(s), dtype=np.float64)
            return
        else:
            V = self.eye - rho*np.outer(s, y)
            self.H = V.dot(self.H).dot(V.T) + rho*np.outer(s, s)

    def get_next_coord(self, coord, force, max_step=0.04):
        """
        获取下一步结构
        """
        # coord = coord.flatten()
        # force = force.flatten()
        if len(self.force) <= 2:
            self.force.append(force)
        else:
            del self.force[0]
            self.force.append(force)
        if self.step_num == 0:
            # 初始化Hessian矩阵
            self.eye = np.eye(force.size, dtype=np.float64)
            self.H = np.eye(force.size, dtype=np.float64)
        else:
            y = self.force[-2] - force
            s = self.step[-1]
            # sy = s.dot(y)
            self.update_hess_BFGS(s, y)
            del self.step[0]
        step = self.H.dot(force)
        # step = self.PRFO_min(coord, force, self.H, self.MAX_STEP)
        step = self.scale_by_max_step(step, self.MAX_STEP)
        self.step.append(step)
        next_coord = coord + step
        # next_coord = next_coord.reshape(int(len(next_coord)/3), 3)
        self.step_num += 1
        return next_coord

    def check_converge(self, coord, force):
        """
        检查是否收敛
        """
        if len(self.step) == 0:
            return False
        # max_force = max(np.linalg.norm(force, axis=1))
        max_force = np.abs(force).max()
        rmsd_force = np.sqrt(np.mean(np.square(force)))
        # max_displace = max(np.linalg.norm(self.step[-1], axis=1))
        max_displace = np.abs(self.step[-1]).max()
        rmsd_displace = np.sqrt(np.mean(np.square(self.step[-1])))
        if (max_force <= self.converge_max_force
                and rmsd_force <= self.converge_rmsd_force
                and max_displace <= self.converge_max_displace
                and rmsd_displace <= self.converge_rmsd_displace):
            return True
        else:
            return False


def atom_align_list(coord_list):
    """
    多分子对齐
    以第一个分子作为参考
    """
    target_coord = coord_list[0]
    aligned_coord_list = [target_coord]
    for coord in coord_list[1:]:
        _, new_coord, _ = atom_align(target_coord, coord)
        aligned_coord_list.append(new_coord)
    return aligned_coord_list


def atom_align_with_index(coord1, coord2, index1=None, index2=None):
    """
    两分子坐标对齐
    coord1: 目标结构
    coord2: 待对齐的结构
    index1 2: 对齐骨架索引
    """
    coord1 = coord1.reshape(-1, 3)
    coord2 = coord2.reshape(-1, 3)
    # 需要对齐的目标原子坐标
    if index1 is not None:
        target_coord1 = coord1[index1]
        target_coord2 = coord2[index2]
    # 中心
    center1 = np.sum(target_coord1, axis=0) / target_coord1.shape[0]
    center2 = np.sum(target_coord2, axis=0) / target_coord2.shape[0]
    # 中心对齐
    target_coord1 = target_coord1 - center1
    target_coord2 = target_coord2 - center2
    P = np.mat(target_coord1)
    Q = np.mat(target_coord2)
    # Kabsch算法
    h = np.matmul(P.T, Q)
    u, s, vt = np.linalg.svd(h)
    v = vt.T
    d = np.linalg.det(np.matmul(v, u.T))
    e = np.array([[1, 0, 0],
                  [0, 1, 0],
                  [0, 0, d]], np.float64)
    r = v @ e @ u.T
    coord3 = np.array(np.matmul(coord2, r), np.float64)
    return (coord1.flatten(), coord3.flatten())


def atom_align(coord1, coord2):
    """
    两分子坐标对齐
    coord1: 目标结构
    coord2: 待对齐的结构
    index1 2: 对齐骨架索引
    """
    coord1 = coord1.reshape(-1, 3)
    coord2 = coord2.reshape(-1, 3)
    # 需要对齐的目标原子坐标
    # target_coord1 = coord1[index1]
    # target_coord2 = coord2[index2]
    target_coord1 = coord1
    target_coord2 = coord2
    # 中心
    center1 = np.sum(target_coord1, axis=0) / target_coord1.shape[0]
    center2 = np.sum(target_coord2, axis=0) / target_coord2.shape[0]
    # 中心对齐
    coord1 = coord1 - center1
    coord2 = coord2 - center2
    # target_coord1 = coord1[index1]
    # target_coord2 = coord2[index2]
    target_coord1 = coord1
    target_coord2 = coord2
    P = np.mat(target_coord1, dtype=np.float64)
    I = np.identity(3, np.float64)
    for i in range(10):
        Q = np.mat(target_coord2, dtype=np.float64)
        # Kabsch算法
        h = np.matmul(P.T, Q)
        u, s, vt = np.linalg.svd(h)
        v = vt.T
        d = np.linalg.det(np.matmul(v, u.T))
        e = np.array([[1, 0, 0],
                      [0, 1, 0],
                      [0, 0, d]], np.float64)
        r = v @ e @ u.T
        coord3 = np.array(np.matmul(target_coord2, r), np.float64)
        if (np.abs(I-r) < 1e-12).all():
            break
        else:
            target_coord2 = coord3
    return (coord1, coord3, r)


def IDPP_interplote(p1, p2, n, method='bfgs'):
    """
    IDPP插值方法
    """
    return IDPP_interplote_BFGS(p1, p2, n)


def IDPP_interplote_BFGS(p1, p2, n):
    """
    IDPP插值方法
    """
    inter = IDPP()
    inter.init_image(p1, p2, n)
    inter.align(0)
    for i in range(n):
        inter.coord_list[i+1][:] = inter.coord_list[i] + (p2 - inter.coord_list[i]) / (n-i+1)
        inter.align(i+1)
        inter.opt_bfgs(i+1)
        if i == 0:
            inter.add_random_offset(i+1)
            inter.align(i+1)
            inter.opt_bfgs(i+1)
            inter.align(i+1)
    r_mat = inter.align(n+1)
    return (inter.coord_list, r_mat)


if __name__ == "__main__":
    '''
    surface_test()
    exit()
    '''
    init_path = 'R1.gjf'
    final_path = 'R2.gjf'
    file_path = r'D:\gaussian\IDPP'
    os.chdir(file_path)
    n = 80
    from Gaussian09 import Gaussian09
    gauss = Gaussian09()
    gauss.init_input(init_path)
    init_coord = gauss.coord
    final_coord = gauss.parse_gjf(final_path)[1]
    init_coord, final_coord, _ = atom_align(init_coord, final_coord)
    # coord_list = IDPP_interplote(init_coord, final_coord, n, method='bfgs_neb')
    coord_list, r_mat = IDPP_interplote(init_coord, final_coord, n, method='bfgs')
    # coord_list, r_mat = IDPP_interplote_no_offset(init_coord, final_coord, n, method='bfgs')
    # coord_list = IDPP_interplote(init_coord, final_coord, n)
    gauss.generate_input(gauss.elements, coord_list[0], os.path.join(file_path, 'inter_' + str(0) + '.gjf'))
    for i in range(n):
        # cord1, cord2 = atom_align(init_coord, coord_list[i])
        gauss.generate_input(gauss.elements, coord_list[i+1], os.path.join(file_path, 'inter_' + str(i+1) + '.gjf'))
    gauss.generate_input(gauss.elements, coord_list[n+1], os.path.join(file_path, 'inter_' + str(n+1) + '.gjf'))
