"""
样条曲线计算
"""

import numpy as np


class Spline():
    def __init__(self):
        """
        三次样条插值法
        """
        self.energy_list = []
        self.grad_list = []
        self.index_list = []
        self.n = 0

    def update_value(self, energy_list, grad_list, index_list, insert_num):
        """
        n最大散点数
        """
        self.energy_list = energy_list
        self.grad_list = grad_list
        self.index_list = index_list
        self.n = insert_num + 2
        self.all_energy = []
        for i in range(len(self.index_list)-1):
            self.all_energy.append(self.energy_list[i])
            index1 = self.index_list[i]
            index2 = self.index_list[i+1]
            e1 = self.energy_list[i]
            e2 = self.energy_list[i+1]
            grad1 = self.grad_list[i]
            grad2 = self.grad_list[i+1]
            point_num = index2 - index1 - 1
            t_list = [(j+1)/(point_num+1) for j in range(point_num)]
            # 中间点能量
            energy_list = self.get_value_list(e1, e2, grad1, grad2, t_list)
            self.all_energy.extend(energy_list)
        self.all_energy.append(self.energy_list[-1])

    def get_value_list(self, x_list, e_list, g_list, t_list):
        """
        三次样条曲线插值
        t_list: 插值坐标列表
        """
        # 构建方程组矩阵
        # a1*x1**3 + b1*x1**2 + c1*x1 + d = f(x1)
        # 3*a1*x1**2 + 2*b1*x1 + c1 = f'(x1)
        # 首先维持数值条件
        f_equation = []
        dim = (len(x_list)-1) * 4
        x_equation_array = np.zeros((dim, dim))
        x_counter = 0
        y_counter = 0
        for i in range(len(e_list)-1):
            x = x_list[i]
            # 每个方程需要满足前后两点的数值条件
            x_equation_array[y_counter, x_counter:x_counter+4] = [x**3, x**2, x, 1]
            f_equation.append(e_list[i])
            x = x_list[i+1]
            y_counter += 1
            x_equation_array[y_counter, x_counter:x_counter+4] = [x**3, x**2, x, 1]
            f_equation.append(e_list[i+1])
            y_counter += 1
            x_counter += 4
        # 维持一阶导数条件
        x_counter = 0
        for i in range(len(g_list)-1):
            x = x_list[i]
            x_equation_array[y_counter, x_counter:x_counter+4] = [3*x**2, 2*x, 1, 0]
            f_equation.append(g_list[i])
            x = x_list[i+1]
            y_counter += 1
            x_equation_array[y_counter, x_counter:x_counter+4] = [3*x**2, 2*x, 1, 0]
            f_equation.append(g_list[i+1])
            y_counter += 1
            x_counter += 4
        solve_result = np.linalg.solve(x_equation_array, np.array([f_equation]).T)
        solve_result = solve_result.T.reshape(-1, 4)
        # 对插值坐标进行分组
        # 分组以中间节点作为分界点
        x_node = x_list[1:-1]
        x_group_list = []
        # 只有存在中间节点时需要分组
        t_list = np.array(t_list)
        x_node = np.array(x_node)
        if len(x_node) > 0:
            for i in range(len(x_node)):
                if i == 0:
                    # 第一个节点前的数据
                    x_group_list.append(t_list[np.where(t_list <= x_node[i])])
                    continue
                # 两节点间的数据
                x_group_list.append(t_list[np.where((t_list <= x_node[i]) & (t_list > x_node[i-1]))])
            # 最后一个节点后的数据
            x_group_list.append(t_list[np.where(t_list > x_node[-1])])
        else:
            x_group_list = [np.array(t_list)]
        # 计算曲线上的点
        y_list = []
        '''
        for item in x_group_list:
            print(item)
        '''
        for i in range(len(x_node)+1):
            # 总共可以有节点数+1个曲线方程构成
            x = x_group_list[i]
            y_list.extend(solve_result[i, 0]*x**3 +
                          solve_result[i, 1]*x**2 +
                          solve_result[i, 2]*x +
                          solve_result[i, 3])
        return y_list


class BezierSpline():
    def __init__(self):
        self.energy_list = []
        self.grad_list = []
        self.index_list = []
        self.n = 0

    def update_value(self, energy_list, grad_list, index_list, insert_num):
        """
        初始化贝塞尔曲线
        n最大散点数
        """
        self.energy_list = energy_list
        self.grad_list = grad_list
        self.index_list = index_list
        self.n = insert_num + 2
        self.all_energy = []
        '''
        for i in range(len(self.index_list)):
            energy_index = self.index_list[i]
            self.all_energy[energy_index] = self.energy_list[i]
        '''
        for i in range(len(self.index_list)-1):
            self.all_energy.append(self.energy_list[i])
            index1 = self.index_list[i]
            index2 = self.index_list[i+1]
            e1 = self.energy_list[i]
            e2 = self.energy_list[i+1]
            grad1 = self.grad_list[i]
            if grad1[0] < 0:
                grad1[0] = -grad1[0]
                grad1[1] = -grad1[1]
            grad2 = self.grad_list[i+1]
            if grad2[0] < 0:
                grad2[0] = -grad2[0]
                grad2[1] = -grad2[1]
            point_num = index2 - index1 - 1
            t_list = [(j+1)/(point_num+1) for j in range(point_num)]
            # 中间点能量
            energy_list = self.get_value_list(e1, e2, grad1, -grad2, t_list)
            self.all_energy.extend(energy_list)
        self.all_energy.append(self.energy_list[-1])

    def get_value_list(self, e1, e2, vec1, vec2, t_list):
        """
        贝塞尔插值
        t: 0~1的小数字列表
        """
        energy_list = []
        for t in t_list:
            # energy_list.append((1-t)*e1 + t*(1-t)*vec1 + t*e2 + t*(1-t)*vec2)
            '''
            l1_start = e1 + vec1 * t
            l1_end = e1 + vec1 + (e2 + vec2 - e1 - vec1) * t
            l2_start = l1_end
            l2_end = e2 + (1-t)*vec2
            n1_start = l1_start + (l1_end-l1_start)*t
            n1_end = l2_start + (l2_end-l2_start)*t
            n = n1_start + (n1_end-n1_start)*t
            energy_list.append(n)
            '''
            p0 = e1
            p1 = e1 + vec1
            p2 = e2 + vec2
            p3 = e2
            b0 = (1-t)**3
            b1 = 3*t*(1-t)**2
            b2 = 3*(1-t)*t**2
            b3 = t**3
            n = b0*p0+b1*p1+b2*p2+b3*p3
            energy_list.append(n)
        return energy_list


def show_spline(x_list, y_list):
    import matplotlib.pyplot as plt
    plt.plot(x_list, y_list)
    plt.show()


if __name__ == "__main__":
    # 三次样条曲线
    x_list = [0, 12, 24, 36, 49, 74, 99]
    x_list = [0, 12, 24, 49, 74, 99]
    y_list = [-296.320123764, -296.149121864, -296.154888245, -296.262627125, -296.335588538, -296.37691617]
    g_list = [0.03431727049060638, 0.017063155105954738, -0.01400914129538355, -0.013849611876686145, -0.007805238526747565, -0.0016631686899285952]
    t_list = [0, 0.12292398170243064, 0.22267599593948095, 0.31786178876644405, 0.4065690071885385, 0.4935342499807594, 0.5778985509395411, 0.660385348277349, 0.7421102916931321, 0.8228532534568791, 0.9027767525143309, 0.9820846924169554, 1.0607787021040267, 1.1390022957355672, 1.216864853787658, 1.2944903831095143, 1.371626538968683, 1.448602866242711, 1.5255231949374923, 1.601749128614206, 1.6784831502099198, 1.7547513205269274, 1.8307098755153097, 1.9065048487736187, 1.9818894336825141, 2.057743756253961, 2.13339564177622, 2.209344699386669, 2.2841420763504456, 2.360199453865108, 2.4347207617831197, 2.509602085370533, 2.584887333812147, 2.6597732551424804, 2.7348462979932884, 2.810305840032854, 2.88579788663607, 2.959981654733277, 3.035033803294996, 3.1094704560396313, 3.1850045534808307, 3.2593071112716814, 3.334179035761811, 3.40843825037355, 3.4834760424471414, 3.559071610751899, 3.6322706934548257, 3.7080802988079515, 3.78500555830352, 3.857588426917689, 3.936470195632792, 4.010270171657014, 4.086433198928669, 4.1604778236787086, 4.238276531893576, 4.314397085636987, 4.3846358163530965, 4.464694534825402, 4.538976057508374, 4.613455738168144, 4.691003923069801, 4.7647345789072535, 4.842547222016289, 4.921779721162424, 4.995619765171396, 5.075471797232967, 5.154225696865003, 5.226504281575754, 5.305868506024579, 5.390279949923814, 5.465536468958078, 5.544769905321851, 5.6309875894869, 5.706056666384366, 5.78656197272616, 5.865955148189464, 5.945693718252346, 6.026499617147572, 6.101489488370068, 6.184380530404888, 6.268174657558506, 6.348877720671968, 6.428094110384145, 6.513588118486841, 6.599464081430137, 6.692107566542773, 6.7711124724951395, 6.858843290665848, 6.9449916472948106, 7.035053842528204, 7.124156377684427, 7.215526463433455, 7.312593670535112, 7.410067925259511, 7.518814511087049, 7.6093058222166725, 7.730510095554649, 7.844694834459304, 7.975860049736922, 8.119548354776667]
    # x_list = [0, 10, 20]
    print(len(t_list))
    x_list = list(np.array(t_list)[x_list])
    # t_list = np.linspace(x_list[0], x_list[-1], 100)
    sp = Spline()
    result1 = sp.get_value_list(x_list, [item*10000 for item in y_list], g_list, t_list)
    # result2 = sp.get_value_list(x_list1, y_list1, g_list1, t_list)
    print(result1.index(max(result1)))
    # print(max(result2))
    show_spline(t_list, result1)
    '''
    # 贝塞尔曲线
    p1 = np.array([0, 0], 'f')
    p2 = np.array([10, 0], 'f')
    p3 = np.array([20, 0], 'f')
    x1 = 1
    y1 = x1 +1
    x2 = 1
    y2 = x2 +1
    x3 = 1
    y3 = x3 +1
    v1 = np.array([x1, y1], 'f')
    v2 = np.array([x2, y2], 'f')
    v3 = np.array([x3, y3], 'f')
    n = 100
    sp = BezierSpline()
    sp.update_value([p1, p2, p3], [v1, v2, v3], [0, int(n/2), n+1], n)
    x_list = []
    y_list = []
    for item in sp.all_energy:
        x_list.append(item[0])
        y_list.append(item[1])
    show_spline(x_list, y_list)
    '''