"""
由共价半径判断分子中成键的原子
"""

import os
import numpy as np


class BondConnect():
    def __init__(self):
        self.bond_table()

    def bond_table(self):
        # 原子共价半径，单位为埃
        self.radius = {
            'X': 0.529,
            'H': 0.4, 'He': 0.28, 'Li': 1.28, 'Be': 0.96, 'B': 0.84, 'C': 0.76,
            'N': 0.71, 'O': 0.66, 'F': 0.57, 'Ne': 0.58, 'Na': 1.66, 'Mg': 1.41,
            'Al': 1.21, 'Si': 1.11, 'P': 1.07, 'S': 1.05, 'Cl': 1.02, 'Ar': 1.06,
            'K': 2.03, 'Ca': 1.76, 'Sc': 1.7, 'Ti': 1.6, 'V': 1.53, 'Cr': 1.39,
            'Mn': 1.39, 'Fe': 1.32, 'Co': 1.26, 'Ni': 1.24, 'Cu': 1.32, 'Zn': 1.22,
            'Ga': 1.22, 'Ge': 1.2, 'As': 1.19, 'Se': 1.2, 'Br': 1.2, 'Kr': 1.16,
            'Rb': 2.2, 'Sr': 1.95, 'Y': 1.9, 'Zr': 1.75, 'Nb': 1.64, 'Mo': 1.54,
            'Tc': 1.47, 'Ru': 1.46, 'Rh': 1.42, 'Pd': 1.39, 'Ag': 1.45, 'Cd': 1.44,
            'In': 1.42, 'Sn': 1.39, 'Sb': 1.39, 'Te': 1.38, 'I': 1.39, 'Xe': 1.4,
            'Cs': 2.44, 'Ba': 2.15, 'La': 2.07, 'Ce': 2.04, 'Pr': 2.03, 'Nd': 2.01,
            'Pm': 1.99, 'Sm': 1.98, 'Eu': 1.98, 'Gd': 1.96, 'Tb': 1.94, 'Dy': 1.92,
            'Ho': 1.92, 'Er': 1.89, 'Tm': 1.9, 'Yb': 1.87, 'Lu': 1.87, 'Hf': 1.75,
            'Ta': 1.7, 'W': 1.62, 'Re': 1.51, 'Os': 1.44, 'Ir': 1.41, 'Pt': 1.36,
            'Au': 1.36, 'Hg': 1.32, 'Tl': 1.45, 'Pb': 1.46, 'Bi': 1.48, 'Po': 1.4,
            'At': 1.5, 'Rn': 1.5, 'Fr': 2.6, 'Ra': 2.21, 'Ac': 2.15, 'Th': 2.06,
            'Pa': 2.0, 'U': 1.96, 'Np': 1.9, 'Pu': 1.87, 'Am': 1.8, 'Cm': 1.69
        }
        # 元素名
        self.elements = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne',
                         'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca',
                         'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
                         'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr',
                         'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn',
                         'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd',
                         'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb',
                         'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg',
                         'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th',
                         'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm',
                         'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Uun',
                         'Uuu', 'Uub']

    def generate_distance_matrix(self, coord):
        """
        生成距离矩阵
        """
        coord = coord.reshape(-1, 3)
        num = coord.shape[0]
        dis_mat = np.zeros((num, num), 'f')
        tmp_coord = np.zeros(coord.shape, 'f')
        for i in range(num):
            # 计算距离
            tmp_coord = coord - coord[i]
            dis_mat[i] = np.linalg.norm(tmp_coord, axis=1)
            dis_mat[i, i] = 0
        return dis_mat

    def generate_bond_length_matrix(self, element):
        """
        生成最大键长度矩阵
        元素表使用原子序数列表
        """
        num = len(element)
        length_matrix = np.zeros((num, num), 'f')
        radius_list = np.array([self.radius[ele.title()] for ele in element], 'f')
        for i in range(num):
            length_matrix[i] = (radius_list + radius_list[i]) * 1.3
            # 对角元最大键长为-1,即自己不能与自己成键
            length_matrix[i, i] = -1
        return length_matrix

    def element_trans(self, element):
        """
        将元素转化为原子序数列表
        """
        ele_num_list = []
        for item in element:
            ele_num_list.append(self.elements.index(item.capitalize()))
        return ele_num_list

    def generate_bond(self, element, coord, fullconnect=False):
        """
        由分子结构生成键连接
        element: 原子序数列表
        coord: 坐标, np.array类型
        fullconnect: 是否强制连接所有原子
        """
        # 距离矩阵
        dis_mat = self.generate_distance_matrix(coord)
        # 成键距离矩阵
        len_mat = self.generate_bond_length_matrix(element)
        # 键矩阵
        bond_mat = np.where(dis_mat <= len_mat, True, False)
        if fullconnect:
            # 保证分子所有原子都被连接到一块
            # 原子组列表,包含了能与1号原子连通的所有原子序号
            bond_mat = bond_mat | bond_mat.T
            a = np.array(bond_mat, bool) | np.identity(bond_mat.shape[0], bool)
            # 待遍历的列表
            while True:
                # 矩阵乘法计算可达矩阵
                last_mat = np.array(a)
                for i in range(a.shape[0]):
                    a = np.dot(a, a.T)
                    if (last_mat == a).all():
                        break
                    else:
                        last_mat[:] = a
                # 可达矩阵计算完毕
                # 判断第一个原子的连通性
                if (a[0]).all():
                    # 全连通
                    break
                else:
                    # 存在未连通情况
                    # 取未连通原子间的距离矩阵
                    tmp_mat = np.array(dis_mat)
                    tmp_mat = tmp_mat + np.diag([np.max(tmp_mat)*2 for i in range(tmp_mat.shape[0])])
                    min_dis = np.min(dis_mat[~a])
                    min_dis_position = np.where(dis_mat == min_dis)
                    x = min_dis_position[0][0]
                    y = min_dis_position[1][0]
                    # 最短连接处添加一个键
                    a[x, y] = True
                    bond_mat[x, y] = True
                    a = a | a.T
                    bond_mat = bond_mat | bond_mat.T
        # 将键连接矩阵转化为邻接表
        bond_mat = np.triu(bond_mat, k=1)
        num = bond_mat.shape[0]
        bond_table = []
        for i in range(num):
            for j in range(num):
                if bond_mat[i, j]:
                    bond_table.append([i, j])
        return bond_table

    def merge_connect(self, connect_1, connect_2):
        """
        将两个连接表合并
        """
        merged_connect = connect_1 + connect_2
        new_connect = []
        for con in merged_connect:
            # 排序,保证后者不小于前者
            if con[-1] < con[0]:
                con = [con[-1], con[0]]
            if con in new_connect:
                continue
            else:
                new_connect.append(con)
        return new_connect

    def check_coonnection_bond_table(self, connect, atom_num):
        """
        检查连接表是否是全连接
        """
        bond_mat = np.zeros((atom_num, atom_num), bool)
        for item in connect:
            bond_mat[item[0], item[1]] = True
        bond_mat = bond_mat | bond_mat.T
        a = np.array(bond_mat, bool) | np.identity(bond_mat.shape[0], bool)
        # 矩阵乘法计算可达矩阵
        last_mat = np.array(a)
        for i in range(a.shape[0]):
            a = np.dot(a, a.T)
            if (last_mat == a).all():
                break
            else:
                last_mat[:] = a
        # 可达矩阵计算完毕
        # 判断第一个原子的连通性
        if (a[0]).all():
            # 全连通
            return True
        else:
            return False

    def check_coonnection_bond_mat(self, bond_mat):
        """
        检查连接表是否是全连接
        """
        bond_mat = bond_mat | bond_mat.T
        a = np.array(bond_mat, bool) | np.identity(bond_mat.shape[0], bool)
        # 矩阵乘法计算可达矩阵
        last_mat = np.array(a)
        for i in range(a.shape[0]):
            a = np.dot(a, a.T)
            if (last_mat == a).all():
                break
            else:
                last_mat[:] = a
        # 可达矩阵计算完毕
        # 判断第一个原子的连通性
        if (a[0]).all():
            # 全连通
            return True
        else:
            return False

    def generate_bond_two_molecule(self, element, coord1, coord2):
        """
        生成两个分子的共同连接方式
        """
        # 距离矩阵
        dis_mat1 = self.generate_distance_matrix(coord1)
        dis_mat2 = self.generate_distance_matrix(coord2)
        # 成键距离矩阵
        len_mat = self.generate_bond_length_matrix(element)
        # 键矩阵
        bond_mat1 = np.where(dis_mat1 <= len_mat, True, False)
        bond_mat2 = np.where(dis_mat2 <= len_mat, True, False)
        # 键矩阵取并集
        bond_mat = bond_mat1 | bond_mat2
        # 判断是否全连接
        while not self.check_coonnection_bond_mat(bond_mat):
            a = np.array(bond_mat) | np.identity(bond_mat.shape[0], bool)
            last_mat = np.array(a)
            for i in range(a.shape[0]):
                a = np.dot(a, a.T)
                if (last_mat == a).all():
                    break
                else:
                    last_mat[:] = a
            # 尝试添加键
            # 选择距离最近的原子对
            tmp_mat1 = np.array(dis_mat1)
            tmp_mat2 = np.array(dis_mat2)
            tmp_mat1 = tmp_mat1 + np.diag([np.max(tmp_mat1)*2 for i in range(tmp_mat1.shape[0])])
            tmp_mat2 = tmp_mat2 + np.diag([np.max(tmp_mat2)*2 for i in range(tmp_mat2.shape[0])])
            min_dis1 = np.min(dis_mat1[~a])
            min_dis2 = np.min(dis_mat2[~a])
            if min_dis1 < min_dis2:
                min_dis_position = np.where(dis_mat1 == min_dis1)
            else:
                min_dis_position = np.where(dis_mat2 == min_dis2)
            x = min_dis_position[0][0]
            y = min_dis_position[1][0]
            a[x, y] = True
            a = a | a.T
            bond_mat[x, y] = True
            bond_mat = bond_mat | bond_mat.T
        # 将键连接矩阵转化为邻接表
        bond_mat = np.triu(bond_mat, k=1)
        num = bond_mat.shape[0]
        bond_table = []
        for i in range(num):
            for j in range(num):
                if bond_mat[i, j]:
                    bond_table.append([i, j])
        return bond_table


if __name__ == "__main__":
    BC = BondConnect()
    data = ['C                  0.18413002    1.25352350    0.32249616',
            'H                  0.64932437    0.31486442    0.54024221',
            'H                  0.64933921    1.53427062    1.24426698',
            'H                  0.54080286    1.75792169   -0.55115535',
            'H                 -1.07586998    1.25353903    0.32249616',
            'H                 -10.07586998    1.25353903    0.32249616',
            'H                 -10.17586998    1.25353903    0.32249616']
    elements = []
    coord = []
    for line in data:
        datatxt = line.strip().split()
        elements.append(datatxt[0])
        coord.append([float(item) for item in datatxt[1:]])
    # elem = BC.element_trans(elements)
    coord = np.array(coord)
    bond = BC.generate_bond(elements, coord)
    print(bond)
    a = np.array([[0, 1, 0, 0, 0, 0],
                  [0, 0, 0, 0, 0, 0],
                  [0, 0, 0, 1, 0, 0],
                  [0, 0, 0, 0, 1, 0],
                  [0, 0, 0, 0, 0, 0],
                  [0, 0, 0, 0, 0, 0]], bool)
    a = a | a.T
    bond_mat = a
    # 一步可及矩阵
    a = a | np.identity(a.shape[0], bool)
    last_mat = np.array(a)
    for i in range(a.shape[0]):
        a = np.dot(a, a)
        if (last_mat == a).all():
            break
        else:
            last_mat[:] = a
    a[0, 3] = False
    a = a | a.T
    a = np.dot(a, a)
    a = np.dot(a, a)
