"""
Version: 0.2
Author: Liu Yangqiu
Date: 2023/2/8
Description: Seaching transition stats with improved Elastic Image Pair (i-EIP) method.
Reference: J. Chem. Theory Comput. 2023, 19 (8), 2410-2417. (doi: 10.1021/acs.jctc.3c00151)
"""

import os
import shutil
import numpy as np
from Gaussian09 import Gaussian09
import IDPP
from IDPP import IDPP_interplote
from redundant_coordinate import RedundantCoordinate
from bond_connect import BondConnect
from Bofill import Bofill_update
from Spline import Spline
from EstimateHessian import EstimateHessian
import perpend_force as PF
import traceback


class ImprovedEIP():
    """
    Improved Elastic Image Pair method for searching transition states.
    """
    def __init__(self):
        self.Bohr2Angs = 0.529177249
        self.BC = BondConnect()
        self.RIC = RedundantCoordinate()
        self.spline = Spline()
        self.calculator = Gaussian09()
        self.estimatehess = EstimateHessian()
        self.step = 0
        self.all_force_counter = 0
        self.hess_counter = 0
        # 像点对目标距离
        self.target_length = 0.0
        # 像点对重生成方向
        self.regeneration_direction = 0
        self.init_iEIP()

    def init_iEIP(self):
        """
        初始化iEIP收敛限等相关数据
        """
        # 异常力计数器,用于判断是否进行像点对重生成
        self.unusual_force_count = 0
        self.unusual_force_max_num = 3
        # 特征向量和像点对向量重叠限度
        self.threshold_overlap = 0.8
        # i-EIP最大计算次数
        self.max_loop = 300
        # 收敛限设置
        self.thresholds = [[0.8, 8e-3],
                           [0.6, 5e-3],
                           [0.5, 1e-3]]
        self.thresholds_level = 0
        self.threshold_length = self.thresholds[self.thresholds_level][0]
        self.threshold_max_perpend_F = self.thresholds[self.thresholds_level][1]
        # 均方根力,用于检查是否更新目标距离
        self.threshold_rms_f_param = 1.2e-2
        self.threshold_rms_f = self.threshold_rms_f_param
        # 文件所在目录
        self.path = ''
        # 初始像点对生成时插值路径计算点数
        self.interpolate_number = 5

    def init_read_structure(self, path1, path2):
        """
        从gjf文件中读取数据
        """
        print('                       ########################')
        print('                       ##     Read input     ##')
        print('                       ########################')
        self.name_tag = os.path.split(path1)[1].split('.')[0].split('_')[0]
        self.path = os.path.split(path1)[0]
        print('Current directory:', self.path)
        print('Input files:', os.path.split(path1)[1], os.path.split(path2)[1])
        self.calculator.init_input(path1)
        # 从两个输入文件中读取结构作为初始结构
        point_A = self.calculator.read_coord_from_input(path1).reshape(-1, 3)
        point_B = self.calculator.read_coord_from_input(path2).reshape(-1, 3)
        # 中心点
        center_A = np.sum(point_A, axis=0) / len(point_A)
        center_B = np.sum(point_B, axis=0) / len(point_B)
        # 将中心点平移到原点
        self.point_A = point_A - center_A
        self.point_B = point_B - center_B
        # 读取元素
        self.elements = self.calculator.elements
        # 笛卡尔坐标的相关处理
        self.point_A = self.point_A.flatten()
        self.point_B = self.point_B.flatten()
        self.num = len(self.elements)
        print('Number of atoms:', self.num)
        print('')

    def set_convergence_threshold(self):
        """
        提高iEIP方法收敛限
        """
        self.thresholds_level += 1
        if self.thresholds_level >= len(self.thresholds):
            self.thresholds_level = len(self.thresholds) - 1
        self.threshold_length = self.thresholds[self.thresholds_level][0]
        self.threshold_max_perpend_F = self.thresholds[self.thresholds_level][1]
        # 均方根力,用于检查是否更新目标距离
        self.threshold_rms_f_param = 1.2e-2
        self.threshold_rms_f = self.threshold_rms_f_param

    def print_info(self):
        """
        打印软件信息
        """
        info_list = []
        info_list.append('|--------------------------------------------------------------------|')
        info_list.append('|                                                                    |')
        info_list.append('|                   Improved Elastic Image Pair method               |')
        info_list.append('|                                                                    |')
        info_list.append('|         If this program benefitted your research please cite:      |')
        info_list.append('|                https://doi.org/10.1021/acs.jctc.3c00151            |')
        info_list.append('|                                                                    |')
        info_list.append('|--------------------------------------------------------------------|')
        info_list.append('')
        info_list.append('')
        for line in info_list:
            print(line)

    def get_force_gauss(self, coord, job_type='A'):
        """
        生成gjf文件,并执行计算,从文件中读取受力
        会尝试读取上一步的chk文件中的SCF信息作为该步波函数初猜
        """
        filename = 'force_' + job_type + '_' + str(self.step) + '.gjf'
        out_filename = 'force_' + job_type + '_' + str(self.step) + '.log'
        fchk_filename = filename[:-3] + 'fchk'
        # 检查文件是否已经计算过
        if os.path.exists(out_filename) and os.path.exists(fchk_filename):
            if self.calculator.check_termnation(out_filename):
                force = self.calculator.get_force_fchk(fchk_filename)
                energy = self.calculator.get_energy(out_filename)
                self.all_force_counter += 1
                return (force, energy)
        # 检查输入文件是否存在
        if not os.path.exists(filename):
            # 检查是否存在可以读取波函数的chk文件,如果有,则尝试读取
            if job_type == 'A' or job_type == 'B':
                oldchk_filename = 'force_' + job_type + '_' + str(self.step-1) + '.chk'
                if os.path.exists(oldchk_filename):
                    self.calculator.generate_input(self.elements, coord, filename, oldchk=oldchk_filename)
                else:
                    self.calculator.generate_input(self.elements, coord, filename, oldchk=None)
            else:
                self.calculator.generate_input(self.elements, coord, filename, oldchk=None)
        self.calculator.run(filename)
        # 如果没有正常结束
        assert self.calculator.check_termnation(out_filename)
        # Hartree/angstorm单位受力
        force = self.calculator.get_force_fchk(fchk_filename)
        # force = self.calculator.get_force(out_filename)
        # Hartree单位能量
        energy = self.calculator.get_energy(out_filename)
        self.all_force_counter += 1
        return (force, energy)

    def get_hess_gauss(self, coord, job_type='A'):
        """
        生成gjf文件,并执行计算,从文件中读取Hessian矩阵和受力
        """
        filename = 'freq_' + job_type + '_' + str(self.hess_counter+1) + '.gjf'
        out_filename = 'freq_' + job_type + '_' + str(self.hess_counter+1) + '.log'
        fchk_filename = filename[:-3] + 'fchk'
        # 检查文件是否已经计算过
        if os.path.exists(out_filename) and os.path.exists(fchk_filename):
            if self.calculator.check_termnation(out_filename):
                hess = self.calculator.get_hess_fchk(fchk_filename)
                force = self.calculator.get_force_fchk(fchk_filename)
                energy = self.calculator.get_energy(out_filename)
                self.all_force_counter += 1
                self.hess_counter += 1
                return (force, energy, hess, filename)
        # 检查输入文件是否存在
        if not os.path.exists(filename):
            oldchk_filename = 'force_A_' + str(self.step-1) + '.chk'
            if os.path.exists(oldchk_filename):
                self.calculator.generate_input(self.elements, coord, filename, type='freq', oldchk=oldchk_filename)
            else:
                self.calculator.generate_input(self.elements, coord, filename, type='freq', oldchk=None)
        self.calculator.run(filename)
        # 如果没有正常结束
        assert self.calculator.check_termnation(out_filename)
        # Hartree/angstorm单位受力
        force = self.calculator.get_force_fchk(fchk_filename)
        hess = self.calculator.get_hess_fchk(fchk_filename)
        # force = self.calculator.get_force(out_filename)
        # Hartree单位能量
        energy = self.calculator.get_energy(out_filename)
        self.all_force_counter += 1
        self.hess_counter += 1
        return (force, energy, hess, filename)

    def get_tangent_direction(self, i, coord_list):
        """
        获取i点处的切线向量
        """
        if i == 0:
            direction = coord_list[i+1] - coord_list[i]
            direction = direction / np.linalg.norm(direction)
        elif i == len(coord_list)-1:
            direction = coord_list[i] - coord_list[i-1]
            direction = direction / np.linalg.norm(direction)
        else:
            dir1 = coord_list[i] - coord_list[i-1]
            dir1 = dir1 / np.linalg.norm(dir1)
            dir2 = coord_list[i+1] - coord_list[i]
            dir2 = dir2 / np.linalg.norm(dir2)
            direction = dir1 + dir2
            direction = direction / np.linalg.norm(direction)
        return direction

    def search_max(self, p1, p2, max_n=200, point_num=1):
        """
        搜索p1 p2间最大值所在区间
        使用样条曲线方案
        但是只考虑只对能量最高点的两侧做插值搜索处理，其它区间不再考虑
        最高点两侧进行插值拟合
        """
        print('                     ###########################')
        print('                     ##  Generate image pair  ##')
        print('                     ###########################')
        p1_, p2_ = self.atom_align(p1, p2)
        p1_ = p1_.reshape(-1, 3)
        p2_ = p2_.reshape(-1, 3)
        # 开始执行插值操作
        print('Node number:', max_n)
        print('Running IDPP interpolation...')
        inter_list, _ = IDPP_interplote(p1_, p2_, max_n)
        print('Interpolation finished')
        inter_list = [item.flatten() for item in inter_list]
        # 计算IDPP插值路径上各点的距离坐标
        distance_list = [0]
        distance_list.extend([np.linalg.norm(inter_list[i]-inter_list[i-1]) for i in range(1, len(inter_list))])
        path_coord = []
        tmp_value = 0
        for i in range(len(distance_list)):
            tmp_value += distance_list[i]
            path_coord.append(tmp_value)
        # IDPP路径总长度
        total_length = path_coord[-1]
        print('Path length: %.4f' % total_length)
        # 对插值结构进行计算
        energy_list = []
        force_list = []
        x_list = []
        g_list = []
        init_max_n = self.interpolate_number
        # init_max_n = 20
        print('')
        print('cycle'.rjust(6), 'index'.rjust(6), 'E/a.u.')
        for i in range(init_max_n):
            # 对路径上的部分点进行计算
            self.step += 1
            node_index = int((i)/(init_max_n-1)*(max_n+1))
            f, e = self.get_force_gauss(inter_list[node_index], 'M')
            # 记录初始点和终点
            if i == 0:
                self.point_init = inter_list[node_index] / self.Bohr2Angs
                self.energy_init = e
                self.force_init = f
            elif i == (init_max_n-1):
                self.point_final = inter_list[node_index] / self.Bohr2Angs
                self.energy_final = e
                self.force_final = f
            print(str(self.step).rjust(6), str(node_index).rjust(6), str(e))
            force_list.append(f.flatten())
            energy_list.append(e)
            x_list.append(node_index)
            direction = self.get_tangent_direction(node_index, inter_list)
            g_list.append(-np.dot(f, direction))
        spline_energy_list = self.spline.get_value_list(np.array(path_coord, np.float64)[x_list], energy_list, g_list, path_coord)
        # 样条曲线计算
        max_index = spline_energy_list.index(max(spline_energy_list))
        if point_num == 1:
            return inter_list[max_index]
        else:
            # 在最高点左右两侧取能量相近的两个点
            # 比较能量直至两个结构距离大于等于总长度的1/5
            spline_energy_list_left = spline_energy_list[:max_index]
            spline_energy_list_left.reverse()
            # 筛选点
            n = 25
            while True:
                p1_index = max_index - n
                p2_index = max_index + n
                if p1_index < 0:
                    p1_index = 0
                if p2_index > (max_n-1):
                    p2_index = max_n - 1
                if p1_index == 0 and p2_index == (max_n-1):
                    print('Index of maximum:', max_index)
                    print('Selected index:', p1_index, p2_index)
                    print('')
                    return (inter_list[p1_index], inter_list[p2_index])
                p1 = inter_list[p1_index]
                p2 = inter_list[p2_index]
                p1, p2 = self.atom_align(p1, p2)
                length = np.linalg.norm(p2 - p1)
                if length < self.threshold_length:
                    n += 1
                else:
                    print('Index of maximum:', max_index)
                    print('Selected index:', p1_index, p2_index)
                    print('')
                    return (inter_list[p1_index], inter_list[p2_index])

    def atom_align(self, coord1, coord2, index1=None, index2=None):
        """
        两分子坐标对齐
        coord1: 目标结构
        coord2: 待对齐的结构
        index1 2: 对齐骨架索引
        """
        coord1 = coord1.reshape(-1, 3)
        coord2 = coord2.reshape(-1, 3)
        # 需要对齐的目标原子坐标
        target_coord1 = coord1
        target_coord2 = coord2
        # 中心
        center1 = np.sum(target_coord1, axis=0) / target_coord1.shape[0]
        center2 = np.sum(target_coord2, axis=0) / target_coord2.shape[0]
        # 中心对齐
        coord1 = coord1 - center1
        coord2 = coord2 - center2
        target_coord1 = coord1
        target_coord2 = coord2
        P = np.mat(target_coord1)
        Q = np.mat(target_coord2)
        # Kabsch算法
        h = np.matmul(P.T, Q)
        u, s, vt = np.linalg.svd(h)
        v = vt.T
        d = np.linalg.det(np.matmul(v, u.T))
        e = np.array([[1, 0, 0],
                     [0, 1, 0],
                     [0, 0, d]], np.float64)
        r = v @ e @ u.T
        coord3 = np.array(np.matmul(coord2, r), np.float64)
        return (coord1.flatten(), coord3.flatten())

    def cart_to_ric(self, elements, p1, p2):
        """
        将两组坐标转化为冗余内坐标
        """
        self.bond_connect = self.BC.generate_bond_two_molecule(elements, p1, p2)
        q1, b_mat1 = self.RIC.cart2ric_coord_prime(p1, self.bond_connect)
        q2, b_mat2 = self.RIC.cart2ric_coord_prime(p2, self.bond_connect)
        return (q1, q2)

    def check_between_two_point(self, p1, p2, e1, e2, f1, f2, max_n=100):
        """
        使用样条曲线方案在坐标列表中搜索最高点
        """
        p1_ = p1.reshape(-1, 3)
        p2_ = p2.reshape(-1, 3)
        f1_ = f1.reshape(-1, 3)
        f2_ = f2.reshape(-1, 3)
        # 只考虑连接方向的受力分力
        f1_tmp = f1_.flatten()
        f2_tmp = f2_.flatten()
        direction = p2_ - p1_
        direction = direction.flatten()
        length = np.linalg.norm(direction)
        direction = direction / length
        fp1 = np.dot(f1_tmp, direction)
        fp2 = np.dot(f2_tmp, direction)
        if fp1 <= 0 and fp2 >= 0:
            # 正常方向
            return 0
        elif fp1 <= 0 and fp2 <= 0:
            return 1
        elif fp1 >= 0 and fp2 >= 0:
            return -1
        # 开始执行插值操作
        inter_list, r_mat_interplote = IDPP_interplote(p1_, p2_, max_n)
        # 受力需要随着结构一起旋转
        f2_ = np.array(np.matmul(f2_, r_mat_interplote))
        # 一维化
        inter_list = [item.flatten() for item in inter_list]
        force_list = [f1_, f2_]
        force_list = [item.flatten() for item in force_list]
        # 计算IDPP插值路径上各点的距离坐标
        distance_list = [0]
        distance_list.extend([np.linalg.norm(inter_list[i]-inter_list[i-1]) for i in range(1, len(inter_list))])
        path_coord = []
        tmp_value = 0
        for i in range(len(distance_list)):
            tmp_value += distance_list[i]
            path_coord.append(tmp_value)
        # 对插值结构进行计算
        energy_list = [e1, e2]
        # 头尾两点
        x_list = [0, len(inter_list)-1]
        g_list = []
        for i in range(len(force_list)):
            direction = self.get_tangent_direction(x_list[i], inter_list)
            g_list.append(-np.dot(force_list[i], direction))
        spline_energy_list = self.spline.get_value_list(np.array(path_coord)[x_list], energy_list, g_list, path_coord)
        # 对最高点进行计算
        max_index = spline_energy_list.index(max(spline_energy_list))
        # 判断最高点是否在插值段中间，或者在两端
        if max_index == 0:
            return -1
        elif max_index > 0 and max_index < (len(inter_list)-1):
            return 0
        else:
            return 1

    def search_max_between_4_points(self, p1, p2, e1, e2, f1, f2, dir, max_n=100):
        """
        使用梯度爬升法搜索最高点并生成新像点对
        """
        # 两点距离计算
        # p1_tmp, p2_tmp = self.atom_align(p1, p2)
        p1_tmp = p1
        p2_tmp = p2
        source_distance = max(np.linalg.norm(p2_tmp - p1_tmp), self.target_length)
        # 重新选择两个插值端点
        if dir == -1:
            p1_ = self.point_init
            p2_ = p1
            p3_ = p2
        elif dir == 1:
            p1_ = p2
            p2_ = self.point_final
            p3_ = p1
        else:
            print('Error on index while search max point between 4 points')
            return
        p1_, p2_, r_mat_align = IDPP.atom_align(p1_, p2_)
        p1_ = p1_.reshape(-1, 3)
        p2_ = p2_.reshape(-1, 3)
        p3_ = p3_.reshape(-1, 3)
        # 像点对的中点
        if dir == -1:
            inter_list, _ = IDPP_interplote(p2_, p3_, max_n)
        else:
            inter_list, _ = IDPP_interplote(p3_, p1_, max_n)
        pm_ = inter_list[int((max_n+2)/2)]
        pm_ = pm_.flatten()
        # 寻找像点对另一个点
        # 开始执行插值操作
        inter_list, r_mat_interplote = IDPP_interplote(p1_, p2_, max_n)
        # 一维化
        inter_list = [item.flatten() for item in inter_list]
        # 计算IDPP插值路径上各点的距离坐标
        if dir == 1:
            p1_tmp = pm_
        else:
            p2_tmp = pm_
            inter_list.reverse()
        last_distance = 0
        # 选择相邻的点作为新EIP点
        max_loop = 3
        for j in range(max_loop):
            for i in range(1, len(inter_list)):
                new_index = min(i, len(inter_list)-1)
                if dir == 1:
                    p2_tmp = inter_list[new_index]
                else:
                    p1_tmp = inter_list[new_index]
                p1_tmp, p2_tmp = self.atom_align(p1_tmp, p2_tmp)
                dis = np.linalg.norm(p2_tmp - p1_tmp)
                new_p1 = p1_tmp
                new_p2 = p2_tmp
                if dis >= source_distance:
                    if j == max_loop-1:
                        # 最后一圈,直接退出
                        return (new_p1, new_p2)
                    else:
                        if (dis - last_distance) < 1e-3:
                            # 距离精度足够,直接退出
                            return (new_p1, new_p2)
                        else:
                            # 尝试高精度再搜索一次
                            # 在该点和左侧点中间再次插值
                            inter_list, r_mat_interplote = IDPP_interplote(inter_list[i-1].reshape(-1, 3), inter_list[i].reshape(-1, 3), max_n)
                            inter_list = [item.flatten() for item in inter_list]
                            last_distance = dis
                            break
                elif i == len(inter_list)-1:
                    # 已经搜索到终点,直接退出
                    return (new_p1, new_p2)

    def generate_ts_file(self, coord):
        """
        生成过渡态的gjf文件
        """
        filename = 'ts_' + self.name_tag + '_' + str(self.all_force_counter) + '.gjf'
        self.calculator.generate_input(self.elements, coord, filename, 'ts')
        return filename

    def print_current_data(self, p1, p2, f1, f2, e1, e2, target_length):
        """
        打印当前状态
        """
        direction = p2 - p1
        direction = direction / np.linalg.norm(direction)
        fp1 = np.dot(f1, direction) * direction
        fp2 = np.dot(f2, direction) * direction
        fv1 = f1 - fp1
        fv2 = f2 - fp2
        # 计算垂直力
        total_fv = np.concatenate([fv1, fv2])
        rms_force = np.sqrt(np.mean(total_fv**2))
        print(str(self.step).rjust(3),
              ('%.2f' % np.linalg.norm(p2-p1)).rjust(4),
              ('%.2f' % target_length).rjust(4),
              ('%.4f' % (np.dot(f1, direction))).rjust(7),
              ('%.4f' % (np.dot(f2, direction))).rjust(7),
              ('%.4f' % (np.abs(total_fv).max())).rjust(6),
              ('%.4f' % rms_force).rjust(6),
              ('%.4f' % (e1 - e2)).rjust(8),
              ('%.4f' % (e1 + e2)).rjust(6),
              str(self.regeneration_direction).rjust(2)
              )
        return False

    def check_ts_converge(self, f, dx):
        """
        检查精确过渡态是否优化完成
        """
        rmsf = np.sqrt(np.mean(f**2))
        max_f = np.abs(f).max()
        if rmsf <= 3e-4 and max_f <= 4.5e-4:
            return True
        else:
            return False

    def check_all_frozen(self, filaname):
        """
        检查计算输出文件
        """
        with open(filaname, 'r') as f:
            filelines = f.readlines()
        if 'All variables have been frozen.' in filelines[-4]:
            return True
        else:
            return False

    def check_rms_force(self, rms_f):
        """
        检查像点对的rms力是否满足收敛限
        """
        if rms_f <= self.threshold_rms_f:
            return True
        else:
            return False

    def check_max_force(self, max_perpend_f):
        """
        检查像点对的垂直力最大分量是否满足收敛
        """
        if max_perpend_f <= self.threshold_max_perpend_F:
            return True
        else:
            return False

    def check_length(self, length):
        """
        检查像点对的距离是否满足收敛限
        """
        if length <= self.threshold_length:
            return True
        else:
            return False

    def update_converged_force(self, length):
        """
        动态更新收敛限
        """
        self.threshold_rms_f = min(max(length/self.threshold_length, 1.0)*self.threshold_rms_f_param,
                                  self.threshold_rms_f_param*2)
        return

    def get_rmsd(self, p1, p2):
        """
        计算两结构的rmsd
        """
        p1, p2 = self.atom_align(p1, p2)
        rmsd = np.sqrt(np.mean((p2-p1)**2))
        return rmsd

    def generate_freeze_triangle_bond_list(self, coord, f, H):
        """
        生成默认冗余内坐标下的三角形内冻结的键长
        """
        # 特征向量分量大于threshold时才被冻结
        threshold = 0.2
        # 将ts结构转化到默认冗余内坐标下
        # 键连接方式使用反应物和产物连接方式的并集
        bond_connect_A = self.BC.generate_bond(self.elements, self.point_init * self.Bohr2Angs)
        bond_connect_B = self.BC.generate_bond(self.elements, self.point_final * self.Bohr2Angs)
        bond_connect = self.BC.merge_connect(bond_connect_A, bond_connect_B)
        # bond_connect = self.BC.generate_bond(self.elements, coord, fullconnect=True)
        q_H, b_mat_H = self.RIC.cart2ric_coord_prime(coord, bond_connect)
        # 将Hessian转换到冗余内坐标下
        f_inter = self.RIC.cart2ric_force(f, b_mat_H)
        H_inter = self.RIC.cart2ric_hess_prime(coord, bond_connect, H, b_mat_H, f_inter)
        eig_val, eig_vec = np.linalg.eigh(H_inter)
        eig_val = np.real(eig_val)
        eig_vec = np.real(eig_vec).T
        # sort_index = np.argsort(eig_val)
        # 冗余内坐标下Hessian的特征值和特征向量
        eig_vec_follow = eig_vec[0]
        bond_index_list, angle_index_list, dihedral_index_list = self.RIC.get_ric_index_list_prime(H, bond_connect)
        total_index_list = bond_index_list + angle_index_list + dihedral_index_list
        freeze_index = []
        tmp_coord = coord.reshape(-1, 3)
        for i in range(len(total_index_list)):
            if abs(eig_vec_follow[i]) >= threshold:
                freeze_index.append(i)
        freeze_index_list = [total_index_list[i] for i in freeze_index]
        # 判断是否有三角形内的角度被固定
        freeze_angle_list = [freeze_index_list[i] for i in range(len(freeze_index_list)) if len(freeze_index_list[i]) == 3]
        freeze_angle_triangle_list = []
        addition_bond_freeze_list = []
        # 判断两个角度是否使用了同一边
        for i in range(len(freeze_angle_list)):
            item = sorted([freeze_angle_list[i][0], freeze_angle_list[i][2]])
            if item in bond_index_list:
                # 该角度在三角形内,则将整个三角形的边长固定,将内部的角度都删除
                # 可能有重复的,这个在后续步骤中消除
                addition_bond_freeze_list.append(sorted([freeze_angle_list[i][0], freeze_angle_list[i][1]]))
                addition_bond_freeze_list.append(sorted([freeze_angle_list[i][1], freeze_angle_list[i][2]]))
                addition_bond_freeze_list.append(sorted([freeze_angle_list[i][0], freeze_angle_list[i][2]]))
                freeze_angle_triangle_list.append(freeze_angle_list[i])
        # 删除三角形内的角度
        for item in freeze_angle_triangle_list:
            if item in freeze_index_list:
                freeze_index_list.remove(item)
        # 添加冻结键
        for item in addition_bond_freeze_list:
            if item not in freeze_index_list:
                freeze_index_list.append(item)
        freeze_index = []
        for item in addition_bond_freeze_list:
            freeze_index.append(total_index_list.index(item))
        freeze_index = sorted(list(set(freeze_index)))
        return freeze_index

    def generate_freeze_list(self, coord, f, H):
        """
        生成冗余内坐标下的冻结变量
        """
        # 特征向量分量大于threshold时才被冻结
        threshold = 0.2
        # 将ts结构转化到冗余内坐标下
        # 键连接方式使用反应物和产物连接方式的并集
        bond_connect_A = self.BC.generate_bond(self.elements, self.point_init * self.Bohr2Angs)
        bond_connect_B = self.BC.generate_bond(self.elements, self.point_final * self.Bohr2Angs)
        bond_connect = self.BC.merge_connect(bond_connect_A, bond_connect_B)
        # bond_connect = self.BC.generate_bond(self.elements, coord, fullconnect=True)
        q_H, b_mat_H = self.RIC.cart2ric_coord(coord, bond_connect)
        # 将Hessian转换到冗余内坐标下
        f_inter = self.RIC.cart2ric_force(f, b_mat_H)
        H_inter = self.RIC.cart2ric_hess(coord, bond_connect, H, b_mat_H, f_inter)
        eig_val, eig_vec = np.linalg.eigh(H_inter)
        eig_val = np.real(eig_val)
        eig_vec = np.real(eig_vec).T
        # sort_index = np.argsort(eig_val)
        # 冗余内坐标下Hessian的特征值和特征向量
        eig_vec_follow = eig_vec[0]
        bond_index_list, angle_index_list, dihedral_index_list = self.RIC.get_ric_index_list(H, bond_connect)
        total_index_list = bond_index_list + angle_index_list + dihedral_index_list
        freeze_index = []
        tmp_coord = coord.reshape(-1, 3)
        for i in range(len(total_index_list)):
            if abs(eig_vec_follow[i]) >= threshold:
                freeze_index.append(i)
        freeze_index.extend(self.generate_freeze_triangle_bond_list(coord, f, H))
        freeze_index = sorted(list(set(freeze_index)))
        freeze_list = []
        '''
        for i in freeze_index:
            # 检查是否为接近180度的键角
            if len(total_index_list[i]) == 3:
                # 键角,检查角度大小
                item = total_index_list[i]
                angle = self.RIC.q_angle(tmp_coord[item[0]], tmp_coord[item[1]], tmp_coord[item[2]]) / np.pi * 180
                if abs(angle) > 170:
                    freeze_list.append('L ' + ' '.join([str(item+1) for item in total_index_list[i]]) + ' -1 B\n')
                else:
                    freeze_list.append(' '.join([str(item+1) for item in total_index_list[i]]) + ' B\n')
            else:
                freeze_list.append(' '.join([str(item+1) for item in total_index_list[i]]) + ' B\n')
        '''
        for i in freeze_index:
            if len(total_index_list[i]) == 3:
                # 键角,检查角度大小
                item = total_index_list[i]
                angle = self.RIC.q_angle(tmp_coord[item[0]], tmp_coord[item[1]], tmp_coord[item[2]]) / np.pi * 180
                if abs(angle) > 170:
                    freeze_list.append('L ' + ' '.join([str(item+1) for item in total_index_list[i]]) + ' -1 F\n')
                else:
                    freeze_list.append(' '.join([str(item+1) for item in total_index_list[i]]) + ' F\n')
            else:
                freeze_list.append(' '.join([str(item+1) for item in total_index_list[i]]) + ' F\n')
        # 计算冻结的自由度
        freeze_b_mat = b_mat_H[freeze_index]
        freeze_G_mat = self.RIC.G_mat(freeze_b_mat)
        eig_val, eig_vec = np.linalg.eig(freeze_G_mat)
        eig_val = np.real(eig_val)
        eig_vec = np.real(eig_vec)
        sort_index = np.argsort(eig_val)
        eig_val = eig_val[sort_index]
        eig_vec = (eig_vec.T)[sort_index]
        counter = 0
        for item in eig_val:
            if abs(item) > 1e-6:
                counter += 1
        return (freeze_list, counter, bond_index_list)

    def ts_opt_with_freq_file(self, coord, freq_filename):
        """
        使用做过频率分析的文件及chk文件进行过渡态搜索
        """
        # 生成ts输入文件和chk文件
        ts_filename = 'ts_' + self.name_tag + '_' + str(self.all_force_counter) + '.gjf'
        self.calculator.generate_input_ts(self.elements, coord, ts_filename, chk=True,
                                          oldchk_filename=freq_filename, read_hess=True)
        self.calculator.run(ts_filename)
        return ts_filename

    def ts_opt_with_constrained_optimization_file(self, cons_opt_filename):
        """
        使用做过限制性优化的文件及chk文件进行过渡态搜索
        笛卡尔坐标从限制性优化文件中读取
        """
        # 生成ts输入文件和chk文件
        ts_filename = 'ts_' + self.name_tag + '_' + str(self.all_force_counter) + '.gjf'
        coord = self.calculator.read_file_log_opt(cons_opt_filename[:-3] + 'log')
        coord = np.array(coord, dtype=np.float64).flatten()
        self.calculator.generate_input_ts(self.elements, coord, ts_filename, chk=True,
                                          oldchk_filename=cons_opt_filename, read_hess=False)
        self.calculator.run(ts_filename)
        return ts_filename

    def ts_opt_modopt_with_freq_file(self, coord, freeze_list, freq_filename):
        """
        使用做过频率分析的文件及chk文件进行限制性优化-过渡态搜索
        """
        # 生成ts输入文件和chk文件
        ts_filename = 'ts_' + self.name_tag + '_' + str(self.all_force_counter) + '.gjf'
        chk_freq_filename = freq_filename[:-3] + 'chk'
        chk_ts_filename = ts_filename[:-3] + 'chk'
        shutil.copy(chk_freq_filename, chk_ts_filename)
        self.calculator.generate_input_modopt_ts(self.elements, coord, ts_filename, freeze_list)
        self.calculator.run(ts_filename)
        return ts_filename

    def constrained_optimization_with_freq_file(self, coord, freeze_list, freq_filename):
        """
        使用做过频率分析的文件及chk文件进行限制性优化

        参数:
        ----------
        coord: 1-D numpy.array
            坐标,单位为Bohr
        freeze_list: list
            冻结变量列表,表中元素为写入到输入文件中的内容
        freq_filename: str
            频率分析的文件名

        返回:
        ----------
        cons_opt_filename: str
            该限制性优化的文件名
        """
        # 生成限制性优化输入文件
        cons_opt_filename = 'constopt_' + self.name_tag + '_' + str(self.all_force_counter) + '.gjf'
        chk_freq_filename = freq_filename[:-3] + 'chk'
        chk_cons_opt_filename = cons_opt_filename[:-3] + 'chk'
        shutil.copy(chk_freq_filename, chk_cons_opt_filename)
        self.calculator.generate_input_constrained_optimization(self.elements, coord, cons_opt_filename, freeze_list)
        self.calculator.run(cons_opt_filename)
        return cons_opt_filename

    def estimate_Hessian(self, coord, force):
        # 估计Hessian
        q, b_mat = self.RIC.cart2ric_coord_prime(coord, self.bond_connect)
        fq = self.RIC.cart2ric_force(force, b_mat)
        try:
            H = self.estimatehess.guess_hess(coord, self.elements, self.bond_connect)
            H = self.RIC.ric2cart_hess_prime(coord, self.bond_connect, H, b_mat, fq)
            if np.isnan(H).any():
                H = np.identity(self.num*3, dtype=np.float64)
        except:
            H = np.identity(self.num*3, dtype=np.float64)
        return H

    def check_energy_maximum_point(self, energy_maximum_direction=0):
        """
        检查能量极高点是否存在于像点对之间,如果连续三次不存在,则返回False

        参数:
        ----------
        energy_maximum_direction: int
            能量极高点在反应物-像点对-产物插值路径上的方向:
            - 0表示在像点对中间;
            - -1表示在反应物侧;
            - 1表示在产物侧
        """
        # 最高点丢失连续发生3次或以上时返回False
        # 直到最高点处于像点对中间
        if energy_maximum_direction == 0:
            self.unusual_force_count = 0
            self.unusual_force_max_num = 3
            return True
        else:
            self.unusual_force_count += 1
        if self.unusual_force_count >= self.unusual_force_max_num:
            self.unusual_force_max_num = 1
            return False
        else:
            return True

    def update_target_length(self, current_length):
        """
        更新像点对的目标距离
        """
        return max(current_length * 0.9, self.threshold_length-0.03)

    def calculate_overlap(self, direction, H):
        """
        计算像点对方向和Hessian矩阵最低特征值对应特征向量方向的重叠值大小

        参数:
        ----------
        direction: 1-D numpy.array
            像点对方向
        H: 2-D numpy.array
            Hessian矩阵
        """
        direction = direction / np.linalg.norm(direction)
        eig_val, eig_vec = np.linalg.eig(H)
        eig_val = np.real(eig_val)
        sort_index = np.argsort(eig_val)
        eig_val = eig_val[sort_index]
        eig_vec = np.real((eig_vec.T)[sort_index])
        overlap = abs(np.dot(eig_vec[0], direction))
        return overlap

    def run_TS_search(self, p, f, H, freq_filename):
        """"
        执行限制性优化-过渡态搜索

        参数:
        ----------
        p: 1-D numpy.array
            过渡态初猜坐标,单位为Bohr
        f: 1-D numpy.array
            受力
        H: 2-D numpy.array
            Hessian矩阵
        freq_filename: str
            初猜结构频率分析文件名
        """
        # 由于i-EIP方法收敛标准太粗糙,结构中可能存在柔性基团没被完全驰豫
        # 所以先尝试进行限制性优化,限制的变量通过Hessian矩阵最低特征向量判断获得
        freeze_list, freeze_freedom, internal_list = self.generate_freeze_list(p * self.Bohr2Angs, f, H)
        internal_list = [' '.join([str(item2+1) for item2 in item]) + ' B\n' for item in internal_list]
        print('Total Hessian:', self.hess_counter)
        print('')
        print('')
        print('                         ###################')
        print('                         ##   Search TS   ##')
        print('                         ###################')
        if freeze_freedom >= (len(p) - 6)/2 or freeze_freedom == 0:
            # 如果有一半的变量被冻结或者没有冻结变量,直接进行过渡态搜索
            print('Running direct TS searching (read Hessian)...')
            self.ts_opt_with_freq_file(p * self.Bohr2Angs, freq_filename)
        else:
            print('Running constrained optimization...')
            # 先做限制性优化
            cons_opt_filename = self.constrained_optimization_with_freq_file(p * self.Bohr2Angs, freeze_list, freq_filename)
            assert self.calculator.check_termnation(cons_opt_filename[:-3] + 'log')
            # 限制性优化完成,开始过渡态搜索
            # 在少数情况下,限制性优化只执行一步就正常结束
            # 可以用先前频率分析的Hessian作为ts搜索的初始Hessian
            # 这样可以节省一次Hessian计算
            step_num = self.calculator.conunt_step_opt(cons_opt_filename[:-3] + 'log')
            print('Constrained optimization step:', step_num)
            if step_num == 1:
                print('Running TS searching (read Hessian)...')
                ts_opt_filename = self.ts_opt_with_freq_file(p * self.Bohr2Angs, freq_filename)
                assert self.calculator.check_termnation(ts_opt_filename[:-3] + 'log')
            else:
                # print('Total Hessian:', self.hess_counter + 1)
                print('Running TS searching (calculate Hessian)...')
                ts_opt_filename = self.ts_opt_with_constrained_optimization_file(cons_opt_filename[:-3] + 'log')
                assert self.calculator.check_termnation(ts_opt_filename[:-3] + 'log')

    def run_iEIP(self):
        """
        使用微走步方式预测下一步结构
        对Hessian特征向量的重叠大小进行检查
        """
        print('                      ##########################')
        print('                      ##  i-EIP optimization  ##')
        print('                      ##########################')
        step_size = 0.3
        p1 = np.array(self.point_A)
        p2 = np.array(self.point_B)
        p1, p2 = self.atom_align(p1, p2)
        self.bond_connect = self.BC.generate_bond_two_molecule(self.elements, p1, p2)
        p1 = p1 / self.Bohr2Angs
        p2 = p2 / self.Bohr2Angs
        f_list_1, f_list_2, p_list_1, p_list_2 = [], [], [], []
        # loop变量用于记录i-EIP优化次数,self.step中还包含了像点对生成的步骤
        loop = 0
        current_length = np.linalg.norm(p1 - p2)
        self.target_length = current_length
        self.update_converged_force(current_length)
        print('Initial length of the image pair: %.4f' % current_length)
        print('Running i-EIP optimization')
        print('Maximum number of cycles:', self.max_loop)
        print('')
        print('Convergence thresholds (level %d):' % self.thresholds_level)
        print('                L <= ', self.threshold_length)
        print('          max|Fv| <= ', self.threshold_max_perpend_F)
        print('        |overlap| >= ', self.threshold_overlap, '(after i-EIP optimization)')
        print('(All quantities are given in a.u.)')
        print('')
        print('')
        print('cycle'.rjust(5), 'L'.rjust(1), 'Lt'.rjust(4), 'Fp(A)'.rjust(8), 'Fp(B)'.rjust(6),
              'max|Fv|'.rjust(6), 'rms(Fv)'.rjust(7), 'd(E)'.rjust(6), 'sum(E)'.rjust(7), 'ReGen'.rjust(5))
        while self.step <= self.max_loop:
            self.step += 1
            f1, e1 = self.get_force_gauss(p1 * self.Bohr2Angs, 'A')
            f2, e2 = self.get_force_gauss(p2 * self.Bohr2Angs, 'B')
            f_list_1.append(f1)
            f_list_2.append(f2)
            p_list_1.append(p1)
            p_list_2.append(p2)
            direction = (p2 - p1) / np.linalg.norm(p2 - p1)
            # 笛卡尔坐标转冗余内坐标
            self.print_current_data(p1, p2, f1, f2, e1, e2, self.target_length)
            current_length = np.linalg.norm(p1 - p2)
            if loop == 0:
                H1 = self.estimate_Hessian(p1, f1)
                H2 = self.estimate_Hessian(p2, f2)
            else:
                H1 = H1 + Bofill_update(H1, p_list_1[-1]-p_list_1[-2], f_list_1[-2]-f_list_1[-1])
                H2 = H2 + Bofill_update(H2, p_list_2[-1]-p_list_2[-2], f_list_2[-2]-f_list_2[-1])
            # 检查像点对中间是否存在最高点
            max_index = self.check_between_two_point(p1, p2, e1, e2, f1, f2)
            fv1 = f1 - np.dot(f1, direction) * direction
            fv2 = f2 - np.dot(f2, direction) * direction
            total_force = np.concatenate([fv1, fv2])
            if not self.check_energy_maximum_point(max_index):
                self.regeneration_direction = max_index
                p1, p2 = self.search_max_between_4_points(p1, p2, e1, e2, f1, f2, max_index)
                current_length = np.linalg.norm(p2 - p1)
                self.target_length = max(current_length, self.threshold_length-0.03)
                continue
            self.regeneration_direction = 0
            current_force = np.sqrt(np.mean(total_force**2))
            max_force = max(np.abs(total_force))
            if not self.check_rms_force(current_force):
                dx1, dx2 = PF.get_microstep_total(p1, p2, e1, e2, f1, f2, H1, H2, step_size, step_size,
                                                  self.target_length, step_size / 1000)
            elif not (self.check_max_force(max_force) and self.check_length(current_length)):
                # 满足rms受力条件,但是未满足最大垂直力和距离条件
                self.update_converged_force(current_length)
                # 当前距离收敛,对目标距离进行更新
                self.target_length = self.update_target_length(current_length)
                # 重新预测dx走步
                dx1, dx2 = PF.get_microstep_total(p1, p2, e1, e2, f1, f2, H1, H2, step_size, step_size,
                                                  self.target_length, step_size / 1000)
            else:
                # 像点对垂直力最大分量和距离都满足收敛条件
                # 开始检查overlap值
                p = (p1 + p2) / 2
                print('')
                print('Calculating Hessian matrix...')
                f, e, H, freq_filename = self.get_hess_gauss(p * self.Bohr2Angs, 'H')
                overlap = self.calculate_overlap(direction, H)
                print('|overlap| is %.6f' % overlap)
                if overlap >= self.threshold_overlap:
                    # 重叠值够大,说明过渡态初猜是合理的
                    # 开始执行限制性优化和过渡态搜索
                    self.all_force_counter -= 1
                    # 因为计算overlap时的数据会被用于限制性优化(或过渡态搜索)的第一步
                    # 所以该步被统计了两次,这里减1来避免受力计算重复统计
                    print('i-EIP optimization converged!')
                    print('All force call (i-EIP): ', self.all_force_counter)
                    self.run_TS_search(p, f, H, freq_filename)
                    print('!!   Finished   !!')
                    break
                elif self.hess_counter >= len(self.thresholds):
                    # 收敛限在第一次计算Hessian后被提高
                    # 如果计算Hessian次数大于最高收敛限数量,说明已经在最收敛限下完成了优化,尝试进行过渡态搜索
                    print('i-EIP optimization finished (level %d)' % self.thresholds_level)
                    print('All force call (i-EIP): ', self.all_force_counter)
                    self.run_TS_search(p, f, H, freq_filename)
                    print('!!   Finished   !!')
                    break
                else:
                    self.set_convergence_threshold()
                    # 重叠值不够,提高收敛标准并继续i-EIP优化
                    print('Continue i-EIP since overlap is less than %.1f' % self.threshold_overlap)
                    print('Update convergence thresholds (level %d):' % self.thresholds_level)
                    print('                L <= ', self.threshold_length)
                    print('          max|Fv| <= ', self.threshold_max_perpend_F)
                    print('')
                    print('cycle'.rjust(5), 'L'.rjust(1), 'Lt'.rjust(4), 'Fp(A)'.rjust(8), 'Fp(B)'.rjust(6),
                          'max|Fv|'.rjust(6), 'rms(Fv)'.rjust(7), 'd(E)'.rjust(6), 'sum(E)'.rjust(7), 'ReGen'.rjust(5))
                    self.target_length = self.update_target_length(current_length)
                    dx1, dx2 = PF.get_microstep_total(p1, p2, e1, e2, f1, f2, H1, H2, step_size, step_size,
                                                      self.target_length, step_size / 1000)
            p1 = p1 + dx1
            p2 = p2 + dx2
            # 两点距离和能量校正
            loop += 1
        else:
            print('i-EIP optimization failure: exceeding maximum number of cycles.')


if __name__ == "__main__":
    base_path = os.getcwd()
    import sys
    argv_1 = sys.argv[1]
    argv_2 = sys.argv[2]
    os.chdir(base_path)
    path1 = os.path.join(base_path, argv_1)
    path2 = os.path.join(base_path, argv_2)
    try:
        i_EIP = ImprovedEIP()
        i_EIP.print_info()
        i_EIP.init_read_structure(path1, path2)
        i_EIP.point_A, i_EIP.point_B = i_EIP.search_max(i_EIP.point_A, i_EIP.point_B, point_num=2)
        i_EIP.point_A, i_EIP.point_B = i_EIP.atom_align(i_EIP.point_A, i_EIP.point_B)
        i_EIP.run_iEIP()
    except Exception as e:
        print(e)
        traceback.print_exc()
