"""
微迭代走步
控制像点对在垂直方向 能量平衡 距离控制这三个方向上独立走步
"""

import numpy as np


def get_microstep(p1, p2, e1, e2, f1, f2, H1, H2, k1, k2, target_length, step_size):
    """
    计算微小迭代步,使用构造的平行力,三者分别单独走步
    将平移和旋转作用取消掉
    """
    direction = p2 - p1
    length = np.linalg.norm(direction)
    direction = direction / np.linalg.norm(direction)
    fv1 = f1 - np.dot(f1, direction) * direction
    fv2 = f2 - np.dot(f2, direction) * direction
    fv1_val = np.linalg.norm(fv1)
    fv2_val = np.linalg.norm(fv2)
    V = p2 - p1
    # 距离控制力
    k1 = 1
    fL1 = -k1 * -V * (2*(length-target_length)/length)
    fL2 = -k1 * V * (2*(length-target_length)/length)
    fL1_val = np.linalg.norm(fL1)
    fL2_val = np.linalg.norm(fL2)
    if fL1_val > 1e-5:
        fL1 = fL1 / fL1_val * step_size
    if fL2_val > 1e-5:
        fL2 = fL2 / fL2_val * step_size
    # 相对能量控制力
    k2 = 1
    fe = (e2 - e1) / length * direction * k2
    fe_val = np.linalg.norm(fe)
    if fe_val > 1e-5:
        fe = fe / fe_val * min(step_size, fe_val)
    # 垂直力走步
    if fv1_val > 1e-5:
        fv1 = fv1 / fv1_val * step_size
    if fv2_val > 1e-5:
        fv2 = fv2 / fv2_val * step_size
    # 对每个力进行走步
    fv1 = fv1 + fL1 + fe
    fv2 = fv2 + fL2 + fe
    fv1_size = np.linalg.norm(fv1)
    fv2_size = np.linalg.norm(fv2)
    max_size = max(fv1_size, fv2_size)
    if max_size > step_size:
        # 保持两者力比例情况下进行缩小
        fv1 = fv1 / max_size * step_size
        fv2 = fv2 / max_size * step_size
    dx1 = fv1
    dx2 = fv2
    return (dx1, dx2)


def remove_rotation_transition(coord):
    """
    消除笛卡尔坐标下Hessian的平动和转动项
    """
    coord = coord.reshape(-1, 3)
    atom_num = len(coord)
    trans_mat = np.zeros((atom_num*3, 3), np.float64)
    for i in range(atom_num):
        trans_mat[i*3, 0] = 1
        trans_mat[i*3+1, 1] = 1
        trans_mat[i*3+2, 2] = 1
    x = trans_mat[:, 0]
    y = trans_mat[:, 1]
    z = trans_mat[:, 2]
    I_mat = np.array([[np.dot(y, y)+np.dot(z, z), -np.dot(x, y), -np.dot(x, z)],
                      [-np.dot(y, x), np.dot(x, x)+np.dot(z, z), -np.dot(y, z)],
                      [-np.dot(z, x), -np.dot(z, y), np.dot(x, x)+np.dot(y, y)]], np.float64)
    rotate_eig_value, rotate_eig_vector = np.linalg.eigh(I_mat)
    rotate_eig_value = np.real(rotate_eig_value)
    rotate_eig_vector = np.real(rotate_eig_vector)
    I = np.zeros((atom_num*3, 3), np.float64)
    for i in range(len(coord)):
        I[3*i:3*(i+1), 0] = [0, -coord[i][2], coord[i][1]]
        I[3*i:3*(i+1), 1] = [-coord[i][2], 0, coord[i][0]]
        I[3*i:3*(i+1), 2] = [-coord[i][1], coord[i][0], 0]
    # 合并平移-旋转向量
    rot_trans_mat = np.zeros((atom_num*3, 6), np.float64)
    rot_trans_mat[:, :3] = trans_mat[:]
    rot_trans_mat[:, 3:] = I[:]
    # 归一化各向量
    for i in range(rot_trans_mat.shape[1]):
        rot_trans_mat[:, i] = rot_trans_mat[:, i] / np.linalg.norm(rot_trans_mat[:, i])
    return [vec for vec in rot_trans_mat.T]


def orithogonality(vec_list, n=None):
    """
    施密特正交化,以vec_list中的第一个向量作为初始基向量
    """
    # 已正交化的向量列表
    if n is None:
        orith_vec = [vec_list[0]]
        for vec in vec_list[1:]:
            new_vec = vec
            for base_vec in orith_vec:
                new_vec = new_vec - base_vec.dot(new_vec) / base_vec.dot(base_vec) * base_vec
            orith_vec.append(new_vec)
    else:
        orith_vec = [vec_list[i] for i in range(n)]
        for vec in vec_list[n:]:
            new_vec = vec.copy()
            for base_vec in orith_vec:
                new_vec = new_vec - base_vec.dot(new_vec) / base_vec.dot(base_vec) * base_vec
            orith_vec.append(new_vec)
    return orith_vec


def get_microstep_total(p1, p2, e1, e2, f1, f2, H1, H2, step_size1, step_size2, target_length, step_size):
    """
    使用微走步预测下一步结构
    """
    max_microstep = 2000
    source_p1 = np.copy(p1)
    source_p2 = np.copy(p2)
    source_e1 = e1
    source_e2 = e2
    source_f1 = np.copy(f1)
    source_f2 = np.copy(f2)
    k1 = 1
    k2 = 1
    # dx1, dx2 = get_microstep_V2(p1, p2, e1, e2, f1, f2, H1, H2, k1, k2, target_length, step_size)
    # return (dx1, dx2)
    # 将走步的平移项和水平项分离掉
    I1_vec_source = remove_rotation_transition(p1)
    I2_vec_source = remove_rotation_transition(p2)
    I_total_vec_source = remove_rotation_transition(np.concatenate([p1, p2]))
    for i in range(max_microstep):
        dx1, dx2 = get_microstep(p1, p2, e1, e2, f1, f2, H1, H2, k1, k2, target_length, step_size)
        p1 = p1 + dx1
        p2 = p2 + dx2
        # 从整体上消除平移和旋转项目
        total_dx1 = p1 - source_p1
        total_dx2 = p2 - source_p2
        I1_vec = I1_vec_source.copy()
        I2_vec = I2_vec_source.copy()
        I_total_vec = I_total_vec_source.copy()
        I1_vec.append(total_dx1)
        I2_vec.append(total_dx2)
        I_total_vec.append(np.concatenate([total_dx1, total_dx2]))
        orth_dx1 = orithogonality(I1_vec, n=6)
        orth_dx2 = orithogonality(I2_vec, n=6)
        orth_dx_total = orithogonality(I_total_vec, n=6)
        total_dx1 = orth_dx1[-1]
        total_dx2 = orth_dx2[-1]
        total_dx = orth_dx_total[-1]
        p1 = source_p1 + total_dx[:len(p1)]
        p2 = source_p2 + total_dx[len(p1):]
        # 能量和受力估计
        # 判断走步是否收敛或者达到最大步长
        dir1 = p1 - source_p1
        dir2 = p2 - source_p2
        e1 = source_e1 + np.dot(-source_f1, dir1) + np.dot(dir1, H1).dot(dir1) / 2
        e2 = source_e2 + np.dot(-source_f2, dir2) + np.dot(dir2, H2).dot(dir2) / 2
        f1 = source_f1 - np.dot(dir1, H1)
        f2 = source_f2 - np.dot(dir2, H2)
        step1 = np.linalg.norm(dir1)
        step2 = np.linalg.norm(dir2)
        if step1 >= step_size1 and step2 >= step_size2:
            dx1 = (p1 - source_p1) / np.linalg.norm(p1 - source_p1) * step_size1
            dx2 = (p2 - source_p2) / np.linalg.norm(p2 - source_p2) * step_size2
            # print('overchieve:', i)
            return (dx1, dx2)
        if step1 >= step_size1:
            dx1 = (p1 - source_p1) / np.linalg.norm(p1 - source_p1) * step_size1
            p1 = source_p1 + dx1
            # print('step1:', i)
            return (dx1, p2 - source_p2)
        if step2 >= step_size2:
            dx2 = (p2 - source_p2) / np.linalg.norm(p2 - source_p2) * step_size2
            p2 = source_p2 + dx2
            # print('step2:', i)
            return (p1 - source_p1, dx2)
        if i == (max_microstep-1):
            dx1 = (p1 - source_p1)
            dx2 = (p2 - source_p2)
            # print('step total:', i)
            return (dx1, dx2)

