"""
构建冗余内坐标的B矩阵
"""


import numpy as np
import math
import derivatives


class RedundantCoordinate():
    def __init__(self):
        self.svd_inv_thresh = 3.16e-4

    def get_ric_index_list_prime(self, cart_coord, connection):
        """
        获取冗余内坐标下的连接索引表
        参数:
            cart_coord: 笛卡尔坐标值,numpy.array [[x, y, z], [x, y, z]...]
            connection: 连接表,[[i1, i2]...]
        返回:
            [[bond_index], [angle_index], [dihedral_index]]
        """
        # 键连接表转化为邻接表
        cart_coord = cart_coord.reshape(-1, 3)
        connect_table = self.connection_trans(connection)
        # connect_table = connection
        bond_index_list = self.get_bond_list(connect_table)
        angle_index_list = self.get_angle_list_prime(connect_table)
        dihedral_index_list = self.get_dihedral_list(connect_table)
        ric_index_list = [bond_index_list, angle_index_list, dihedral_index_list]
        return ric_index_list

    def get_ric_index_list(self, cart_coord, connection):
        """
        获取冗余内坐标下的连接索引表
        参数:
            cart_coord: 笛卡尔坐标值,numpy.array [[x, y, z], [x, y, z]...]
            connection: 连接表,[[i1, i2]...]
        返回:
            [[bond_index], [angle_index], [dihedral_index]]
        """
        # 键连接表转化为邻接表
        cart_coord = cart_coord.reshape(-1, 3)
        connect_table = self.connection_trans(connection)
        # connect_table = connection
        bond_index_list = self.get_bond_list(connect_table)
        angle_index_list = self.get_angle_list(connect_table)
        dihedral_index_list = self.get_dihedral_list(connect_table)
        ric_index_list = [bond_index_list, angle_index_list, dihedral_index_list]
        return ric_index_list

    def cart2ric_coord_prime(self, cart_coord, connection, output_type=False):
        """
        笛卡尔坐标转冗余内坐标
        参数:
            cart_coord: 笛卡尔坐标值,numpy.array [[x, y, z], [x, y, z]...]
            connection: 连接表,[[i1, i2]...]
            output_type: 是否输出键长键角二面角数量列表
        返回:
            弧坐标值 对应Wilson's B矩阵
        """
        cart_coord = cart_coord.reshape(-1, 3)
        bond_index_list, angle_index_list, dihedral_index_list = self.get_ric_index_list_prime(cart_coord, connection)
        # 弧坐标
        arc_coord = []
        # 弧坐标一阶导
        arc_coord_num = len(bond_index_list) + len(angle_index_list) + len(dihedral_index_list)
        cart_coord_num = len(cart_coord)
        dq_arc_mat = np.zeros((arc_coord_num, cart_coord_num*3), np.float64)
        counter = 0
        # 计算键长项
        for index_list in bond_index_list:
            item = cart_coord[index_list]
            arc_coord.append(self.q_bond(item[0], item[1]))
            # dq_arc_result = self.dq_bond(item[0], item[1])
            dq_arc_result = self.dq_bond(item[0], item[1])
            # print(index_list, dq_arc_result)
            for i, index in enumerate(index_list):
                dq_arc_mat[counter][index*3:(index+1)*3] = dq_arc_result[i*3:(i+1)*3]
            counter += 1
        # 计算键角项
        for index_list in angle_index_list:
            item = cart_coord[index_list]
            arc_coord.append(self.q_angle(item[0], item[1], item[2]))
            dq_arc_result = self.dq_angle(item[0], item[1], item[2])
            for i, index in enumerate(index_list):
                dq_arc_mat[counter][index*3:(index+1)*3] = dq_arc_result[i*3:(i+1)*3]
            counter += 1
        # 计算二面角项
        for index_list in dihedral_index_list:
            item = cart_coord[index_list]
            arc_coord.append(self.q_dihedral(item[0], item[1], item[2], item[3]))
            dq_arc_result = self.dq_dihedral(item[0], item[1], item[2], item[3])
            # print(index_list, dq_arc_result)
            for i, index in enumerate(index_list):
                dq_arc_mat[counter][index*3:(index+1)*3] = dq_arc_result[i*3:(i+1)*3]
            counter += 1
        if output_type:
            return (np.array(arc_coord), dq_arc_mat, [len(bond_index_list), len(angle_index_list), len(dihedral_index_list)])
        else:
            return (np.array(arc_coord), dq_arc_mat)

    def cart2ric_coord(self, cart_coord, connection, output_type=False):
        """
        笛卡尔坐标转冗余内坐标
        参数:
            cart_coord: 笛卡尔坐标值,numpy.array [[x, y, z], [x, y, z]...]
            connection: 连接表,[[i1, i2]...]
            output_type: 是否输出键长键角二面角数量列表
        返回:
            弧坐标值 对应Wilson's B矩阵
        """
        cart_coord = cart_coord.reshape(-1, 3)
        bond_index_list, angle_index_list, dihedral_index_list = self.get_ric_index_list(cart_coord, connection)
        # 弧坐标
        arc_coord = []
        # 弧坐标一阶导
        arc_coord_num = len(bond_index_list) + len(angle_index_list) + len(dihedral_index_list)
        cart_coord_num = len(cart_coord)
        dq_arc_mat = np.zeros((arc_coord_num, cart_coord_num*3), np.float64)
        counter = 0
        # 计算键长项
        for index_list in bond_index_list:
            item = cart_coord[index_list]
            arc_coord.append(self.q_bond(item[0], item[1]))
            # dq_arc_result = self.dq_bond(item[0], item[1])
            dq_arc_result = self.dq_bond(item[0], item[1])
            # print(index_list, dq_arc_result)
            for i, index in enumerate(index_list):
                dq_arc_mat[counter][index*3:(index+1)*3] = dq_arc_result[i*3:(i+1)*3]
            counter += 1
        # 计算键角项
        for index_list in angle_index_list:
            item = cart_coord[index_list]
            arc_coord.append(self.q_angle(item[0], item[1], item[2]))
            dq_arc_result = self.dq_angle(item[0], item[1], item[2])
            for i, index in enumerate(index_list):
                dq_arc_mat[counter][index*3:(index+1)*3] = dq_arc_result[i*3:(i+1)*3]
            counter += 1
        # 计算二面角项
        for index_list in dihedral_index_list:
            item = cart_coord[index_list]
            arc_coord.append(self.q_dihedral(item[0], item[1], item[2], item[3]))
            dq_arc_result = self.dq_dihedral(item[0], item[1], item[2], item[3])
            # print(index_list, dq_arc_result)
            for i, index in enumerate(index_list):
                dq_arc_mat[counter][index*3:(index+1)*3] = dq_arc_result[i*3:(i+1)*3]
            counter += 1
        if output_type:
            return (np.array(arc_coord), dq_arc_mat, [len(bond_index_list), len(angle_index_list), len(dihedral_index_list)])
        else:
            return (np.array(arc_coord), dq_arc_mat)

    def cart2ric_coord_factor(self, cart_coord, connection):
        """
        笛卡尔坐标转冗余内坐标的索引序列
        参数:
            cart_coord: 笛卡尔坐标值,numpy.array [[x, y, z], [x, y, z]...]
            connection: 连接表,[[i1, i2]...]
        返回:
            冗余内坐标下距离计算的弧半径
        """
        cart_coord = cart_coord.reshape(-1, 3)
        bond_index_list, angle_index_list, dihedral_index_list = self.get_ric_index_list(cart_coord, connection)
        # 坐标
        arc_coord = []
        counter = 0
        # 计算键长项
        for index_list in bond_index_list:
            item = cart_coord[index_list]
            arc_coord.append(0.5)
            counter += 1
        # 计算键角项
        for index_list in angle_index_list:
            item = cart_coord[index_list]
            arc_coord.append((self.q_bond(item[0], item[1]) + self.q_bond(item[1], item[2]))/4)
            counter += 1
        # 计算二面角项
        for index_list in dihedral_index_list:
            item = cart_coord[index_list]
            arc_coord.append((self.q_bond(item[0], item[1]) + self.q_bond(item[2], item[3]))/4)
            counter += 1
        return np.array(arc_coord, np.float64)

    def ric2cart_coord(self, base_cart_coord, base_ric_coord, delta_ric_coord, b_mat, connect):
        """
        冗余内坐标转笛卡尔坐标
        参数:
            base_cart_coord: 走步前的笛卡尔坐标
            base_ric_coord: 走步前的冗余内坐标
            delta_ric_coord: 冗余内坐标走步
            b_mat: B矩阵
            connect: 连接表
        返回:
            笛卡尔坐标值
        """
        # 求当前b矩阵的广义逆
        old_ric_coord, b_mat, type_list = self.cart2ric_coord(base_cart_coord, connect, output_type=True)
        # b_mat_inv = np.linalg.pinv(b_mat)
        b_mat_inv = self.Bt_inv(b_mat)
        # b_mat_inv = np.matmul(b_mat.T, self.G_inv(b_mat))
        current_delta_ric = np.copy(delta_ric_coord)
        cart_coord = np.array(base_cart_coord)
        q_target = base_ric_coord + delta_ric_coord
        # 正规化目标坐标
        q_target = self.normal_ric_coord(q_target, type_list)
        max_step = 50
        last_rms = 9999
        for i in range(max_step):
            cart_diff = np.dot(b_mat_inv.T, current_delta_ric)
            # 求对应笛卡尔坐标
            cart_coord += cart_diff
            cart_rms = np.sqrt(np.mean(cart_diff**2))
            # 求更新的笛卡尔坐标对应弧坐标
            ric_coord, new_b_mat = self.cart2ric_coord(cart_coord, connect)
            # 计算新弧坐标的差是否满足收敛条件
            ric_coord = self.correct_dihedral(q_target, ric_coord, type_list)
            # current_delta_ric = delta_ric_coord - (ric_coord - base_ric_coord)
            current_delta_ric = q_target - ric_coord
            # b_mat_inv = np.linalg.pinv(new_b_mat)
            # b_mat_inv = self.B_inv(new_b_mat)
            # b_mat_inv = np.matmul(new_b_mat.T, self.G_inv(new_b_mat))
            if cart_rms < last_rms:
                best_cart = cart_coord.copy()
                last_rms = cart_rms
            elif i != 0:
                cart_coord = best_cart
                print('transform fail, using best cart', i, end=' ')
                break
            else:
                raise Exception('Internal -> Cartesian Failed!')
            if cart_rms < 1e-6:
                # print(np.linalg.norm(cart_diff), end=' ')
                return cart_coord
            elif i == max_step-1:
                print('RIC2CC fail:', cart_rms, end=' ')
                '''
                b_mat_inv = np.matmul(b_mat.T, self.G_inv(b_mat))
                cart_diff = np.dot(b_mat_inv, delta_ric_coord)
                cart_coord = np.array(base_cart_coord) + cart_diff
                '''
                cart_coord = best_cart
                break
        return cart_coord

    def cart2ric_force(self, cart_force, b_mat):
        """
        笛卡尔坐标受力转冗余内坐标下的受力
        """
        # test
        G_inv = self.G_inv(b_mat)
        return np.dot(G_inv, b_mat).dot(cart_force)
        # return self.get_P(b_mat).dot(self.Bt_inv(b_mat).dot(cart_force))
        # return np.dot(self.Bt_inv(b_mat), cart_force)

    def ric2cart_force(self, ric_force, b_mat):
        """
        冗余内坐标下的受力转笛卡尔坐标受力
        """
        # return self.get_P(b_mat).dot(self.Bt_inv(b_mat).dot(cart_force))
        G_mat = b_mat.dot(b_mat.T)
        return self.B_inv(b_mat).dot(G_mat).dot(ric_force)

    def cart2ric_hess_prime(self, cart_coord, connect, cart_hess, b_mat, int_grad):
        """
        笛卡尔坐标下的Hessian矩阵转化为冗余内坐标下的Hessian
        """
        cart_coord = cart_coord.reshape(-1, 3)
        connect_table = self.connection_trans(connect)
        bond_index_list = self.get_bond_list(connect_table)
        angle_index_list = self.get_angle_list_prime(connect_table)
        dihedral_index_list = self.get_dihedral_list(connect_table)
        # 弧坐标
        # 弧坐标一阶导
        cart_coord_num = len(cart_coord)
        # 用于Hessian转换的K矩阵
        K_mat = np.zeros((cart_coord_num*3, cart_coord_num*3), np.float64)
        counter = 0
        # 计算键长项
        for index_list in bond_index_list:
            item = cart_coord[index_list]
            d2q_bond = self.d2q_bond(item[0], item[1])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_bond * int_grad[counter]
            counter += 1
        # 计算键角项
        for index_list in angle_index_list:
            item = cart_coord[index_list]
            d2q_angle = self.d2q_angle(item[0], item[1], item[2])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_angle * int_grad[counter]
            counter += 1
        # 计算二面角项
        for index_list in dihedral_index_list:
            item = cart_coord[index_list]
            d2q_dihedral = self.d2q_dihedral(item[0], item[1], item[2], item[3])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_dihedral * int_grad[counter]
            counter += 1
        # 计算笛卡尔坐标下的Hessian
        # G_inv = self.G_inv(b_mat)
        # ric_hess = G_inv.dot(b_mat).dot(cart_hess-K_mat).dot(b_mat.T).dot(G_inv)
        ric_hess = self.Bt_inv(b_mat).dot(cart_hess-K_mat).dot(self.B_inv(b_mat))
        ric_hess = self.get_P(b_mat).dot(ric_hess).dot(self.get_P(b_mat))
        return ric_hess

    def cart2ric_hess(self, cart_coord, connect, cart_hess, b_mat, int_grad):
        """
        笛卡尔坐标下的Hessian矩阵转化为冗余内坐标下的Hessian
        """
        cart_coord = cart_coord.reshape(-1, 3)
        connect_table = self.connection_trans(connect)
        bond_index_list = self.get_bond_list(connect_table)
        angle_index_list = self.get_angle_list(connect_table)
        dihedral_index_list = self.get_dihedral_list(connect_table)
        # 弧坐标
        # 弧坐标一阶导
        cart_coord_num = len(cart_coord)
        # 用于Hessian转换的K矩阵
        K_mat = np.zeros((cart_coord_num*3, cart_coord_num*3), np.float64)
        counter = 0
        # 计算键长项
        for index_list in bond_index_list:
            item = cart_coord[index_list]
            d2q_bond = self.d2q_bond(item[0], item[1])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_bond * int_grad[counter]
            counter += 1
        # 计算键角项
        for index_list in angle_index_list:
            item = cart_coord[index_list]
            d2q_angle = self.d2q_angle(item[0], item[1], item[2])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_angle * int_grad[counter]
            counter += 1
        # 计算二面角项
        for index_list in dihedral_index_list:
            item = cart_coord[index_list]
            d2q_dihedral = self.d2q_dihedral(item[0], item[1], item[2], item[3])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_dihedral * int_grad[counter]
            counter += 1
        # 计算笛卡尔坐标下的Hessian
        # G_inv = self.G_inv(b_mat)
        # ric_hess = G_inv.dot(b_mat).dot(cart_hess-K_mat).dot(b_mat.T).dot(G_inv)
        ric_hess = self.Bt_inv(b_mat).dot(cart_hess-K_mat).dot(self.B_inv(b_mat))
        ric_hess = self.get_P(b_mat).dot(ric_hess).dot(self.get_P(b_mat))
        return ric_hess

    def ric2cart_hess_prime(self, cart_coord, connect, ric_hess, b_mat, int_grad):
        """
        冗余内坐标下的Hessian转笛卡尔坐标的Hessian
        """
        cart_coord = cart_coord.reshape(-1, 3)
        connect_table = self.connection_trans(connect)
        bond_index_list = self.get_bond_list(connect_table)
        angle_index_list = self.get_angle_list_prime(connect_table)
        dihedral_index_list = self.get_dihedral_list(connect_table)
        # 弧坐标
        # 弧坐标一阶导
        cart_coord_num = len(cart_coord)
        # 用于Hessian转换的K矩阵
        K_mat = np.zeros((cart_coord_num*3, cart_coord_num*3), np.float64)
        counter = 0
        # 计算键长项
        for index_list in bond_index_list:
            item = cart_coord[index_list]
            d2q_bond = self.d2q_bond(item[0], item[1])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_bond * int_grad[counter]
            counter += 1
        K_mat[:] = 0
        # 计算键角项
        for index_list in angle_index_list:
            item = cart_coord[index_list]
            d2q_angle = self.d2q_angle(item[0], item[1], item[2])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_angle * int_grad[counter]
            counter += 1
        # 计算二面角项
        for index_list in dihedral_index_list:
            item = cart_coord[index_list]
            d2q_dihedral = self.d2q_dihedral(item[0], item[1], item[2], item[3])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_dihedral * int_grad[counter]
            counter += 1
        # 计算笛卡尔坐标下的Hessian
        cart_hess = b_mat.T.dot(ric_hess).dot(b_mat) + K_mat
        return cart_hess

    def ric2cart_hess(self, cart_coord, connect, ric_hess, b_mat, int_grad):
        """
        冗余内坐标下的Hessian转笛卡尔坐标的Hessian
        """
        cart_coord = cart_coord.reshape(-1, 3)
        connect_table = self.connection_trans(connect)
        bond_index_list = self.get_bond_list(connect_table)
        angle_index_list = self.get_angle_list(connect_table)
        dihedral_index_list = self.get_dihedral_list(connect_table)
        # 弧坐标
        # 弧坐标一阶导
        cart_coord_num = len(cart_coord)
        # 用于Hessian转换的K矩阵
        K_mat = np.zeros((cart_coord_num*3, cart_coord_num*3), np.float64)
        counter = 0
        # 计算键长项
        for index_list in bond_index_list:
            item = cart_coord[index_list]
            d2q_bond = self.d2q_bond(item[0], item[1])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_bond * int_grad[counter]
            counter += 1
        K_mat[:] = 0
        # 计算键角项
        for index_list in angle_index_list:
            item = cart_coord[index_list]
            d2q_angle = self.d2q_angle(item[0], item[1], item[2])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_angle * int_grad[counter]
            counter += 1
        # 计算二面角项
        for index_list in dihedral_index_list:
            item = cart_coord[index_list]
            d2q_dihedral = self.d2q_dihedral(item[0], item[1], item[2], item[3])
            k_mat_index_list_a = []
            k_mat_index_list_b = []
            new_index_list = []
            for item in index_list:
                new_index_list.extend([item*3, item*3+1, item*3+2])
            list_length = len(new_index_list)
            for j in new_index_list:
                k_mat_index_list_a.extend([j for i in range(list_length)])
                k_mat_index_list_b.extend(new_index_list)
            K_mat[k_mat_index_list_a, k_mat_index_list_b] += d2q_dihedral * int_grad[counter]
            counter += 1
        # 计算笛卡尔坐标下的Hessian
        cart_hess = b_mat.T.dot(ric_hess).dot(b_mat) + K_mat
        return cart_hess

    def ric_diff(self, ric1, ric2, type_list):
        """
        冗余内坐标的差
        type_list: 键长键角二面角数量列表
        """
        new_ric2 = self.correct_dihedral(ric1, ric2, type_list)
        return new_ric2 - ric1

    def correct_dihedral(self, old_ric_coord, new_ric_coord, type_list):
        """
        对冗余内坐标的走步进行校正
        """
        old_dihedral = old_ric_coord[type_list[0]+type_list[1]:]
        new_dihedral = new_ric_coord[type_list[0]+type_list[1]:]
        dihedral_diff = new_dihedral - old_dihedral
        dihedral_correct_list = np.abs(np.abs(dihedral_diff) - 2 * np.pi) < np.pi/2
        correct_dihedral = np.copy(new_dihedral)
        correct_dihedral[dihedral_correct_list] -= 2 * np.pi * np.sign(dihedral_diff[dihedral_correct_list])
        correct_ric = np.copy(new_ric_coord)
        correct_ric[type_list[0]+type_list[1]:] = correct_dihedral
        return correct_ric

    def normal_ric_coord(self, ric_coord, type_list):
        """
        正规化冗余内坐标, 将二面角范围转换到[-π, π]之间
        """
        dihedral = ric_coord[type_list[0]+type_list[1]:]
        for i in range(len(dihedral)):
            if dihedral[i] < -np.pi:
                dihedral[i] += 2 * np.pi
            elif dihedral[i] > np.pi:
                dihedral[i] -= 2 * np.pi
        new_ric_coord = np.copy(ric_coord)
        new_ric_coord[type_list[0]+type_list[1]:] = dihedral
        return new_ric_coord

    def connect_table2matrix(self, connect_table):
        """
        邻接表转化为邻接矩阵
        """
        max_num = len(connect_table) + 1
        maxtrix = np.zeros((max_num, max_num), int)
        # 两原子相连接,则写为1,否则为0,对角线为0
        for i in range(len(connect_table)):
            for j in connect_table[i]:
                maxtrix[i, j] = 1
                maxtrix[j, i] = 1
        return maxtrix

    def get_bond_list(self, connec_table):
        """
        计算键长项
        参数:邻接表
        """
        bond_index_list = []
        for i in range(len(connec_table)):
            for j in connec_table[i]:
                if i < j:
                    bond_index_list.append([i, j])
        return bond_index_list

    def get_angle_list_prime(self, connect_table):
        """
        计算角度项
        """
        matrix = self.connect_table2matrix(connect_table)
        # 以i为中心,判断m-i-n能否合法成键角
        # 遍历顺序为i m n依次从小到大
        # m-i-n中必须保证m-n不成键
        angle_index_list = []
        for i in range(len(matrix)):
            for m in range(len(matrix[i])):
                if matrix[i, m] == 1:
                    for n in range(m+1, len(matrix[i])):
                        if matrix[i, n] == 1 and matrix[m, n] == 0:
                            # m-n不成键,可以形成键角
                            angle_index_list.append([m, i, n])
                        elif matrix[i, n] == 1:
                            # m-n成键,这里也判断成键
                            angle_index_list.append([m, i, n])
        return angle_index_list

    def get_angle_list(self, connect_table):
        """
        计算角度项
        """
        matrix = self.connect_table2matrix(connect_table)
        # 以i为中心,判断m-i-n能否合法成键角
        # 遍历顺序为i m n依次从小到大
        # m-i-n中必须保证m-n不成键
        angle_index_list = []
        for i in range(len(matrix)):
            for m in range(len(matrix[i])):
                if matrix[i, m] == 1:
                    for n in range(m+1, len(matrix[i])):
                        if matrix[i, n] == 1 and matrix[m, n] == 0:
                            # m-n不成键才能形成键角
                            angle_index_list.append([m, i, n])
        return angle_index_list

    def get_dihedral_list(self, connect_table):
        """
        计算二面角项
        """
        # 如果m-i-j-n四个原子通过键连接在一起,则这四个原子至少成一个二面角
        # 遍历顺序为i j m n分别从小到大
        # 在二面角中,i-j这两个原子不能同时与另一个原子相连接
        # 在满足上一条件下,四个原子无法合理构成二面角,则可以构建一个T型二面角
        dihedral_index_list = []
        sorted_list = []
        '''
        matrix = self.connect_table2matrix(connect_table)
        num = len(matrix)
        for i in range(num):
            for j in range(i+1, num):
                if matrix[i, j] == 1:
                    # i-j已经确定,尝试搜索两端原子
                    # 搜索形式为m-i-j-n的二面角
                    # 由于i<j,先尝试搜索i端
                    for m in range(num):
                        if m != j and m != i and matrix[i, m] == 1:
                            # m搜索完成,再搜索j端
                            for n in range(num):
                                if n != i and n != j and n != m and matrix[j, n] == 1:
                                    # n搜索完成
                                    if matrix[j, m] == 1 and matrix[i, n] == 1:
                                        # 如果四个原子全连接或构成两个三角形,则使用2-1-3-4顺序
                                        tmp = sorted([m, i, j, n])
                                        dihedral_index_list.append([tmp[1], tmp[0], tmp[2], tmp[3]])
                                    elif matrix[j, m] == 1:
                                        # 如果m与i j构成三角形
                                        pass
                    # 搜索形式为m[n]-i-j的二面角,例如H2C=O
                    for m in range(num):
                        if m != j and m != i and matrix[i, m] == 1 and matrix[j, m] == 0:
                            # m搜索完成,再搜索j端
                            for n in range(num):
                                if n != i and n != j and n != m and matrix[j, n] == 1 and matrix[i, n] == 0:
                                    # n搜索完成
                                    dihedral_index_list.append([m, i, j, n])
            pass
        '''
        for i in range(len(connect_table)):
            for j in connect_table[i]:
                for index_k, k in enumerate(connect_table[j]):
                    # 三个原子编号必须不同
                    if i == k:
                        continue
                    else:
                        # 二面角有两种可能:
                        # 1: A-B-C-D
                        # 2: A-B-C
                        #      |
                        #      D
                        # 第1种情况
                        for m in connect_table[k]:
                            if i == m or j == m:
                                continue
                            elif [m, k, j, i] in dihedral_index_list:
                                continue
                            elif sorted([i, j, k, m]) in sorted_list:
                                continue
                            else:
                                dihedral_index_list.append([i, j, k, m])
                                sorted_list.append(sorted([i, j, k, m]))
                        # 第2种情况
                        for n in connect_table[j][index_k+1:]:
                            if i == n:
                                continue
                            elif [n, k, j, i] in dihedral_index_list:
                                continue
                            elif sorted([i, j, k, n]) in sorted_list:
                                continue
                            else:
                                dihedral_index_list.append([i, j, k, n])
                                sorted_list.append(sorted([i, j, k, n]))
        return dihedral_index_list

    def connection_trans(self, bond_connect):
        """
        键连接转邻接表
        连接表第i个列表内容为i与该索引原子相连接
        索引从0开始
        """
        total_list = [item for data in bond_connect for item in data]
        atom_num = max(total_list) + 1
        connect_table = [[] for i in range(atom_num)]
        # 构建邻接表
        for item in bond_connect:
            index_1 = item[0]
            index_2 = item[1]
            connect_table[index_1].append(index_2)
            connect_table[index_2].append(index_1)
        # 邻接表排序
        for i in range(len(connect_table)):
            connect_table[i].sort()
        return connect_table

    def q_bond(self, p0, p1):
        """
        半径长
        """
        return math.sqrt(sum([(p0[i]-p1[i])**2 for i in range(len(p0))]))

    def q_angle(self, p0, p1, p2):
        """
        角度项
        """
        x0, y0, z0 = p0
        x1, y1, z1 = p1
        x2, y2, z2 = p2
        dx0 = x0 - x1
        dy0 = y0 - y1
        dz0 = z0 - z1
        dx1 = x2 - x1
        dy1 = y2 - y1
        dz1 = z2 - z1
        r0 = np.sqrt(dx0**2 + dy0**2 + dz0**2)
        r1 = np.sqrt(dx1**2 + dy1**2 + dz1**2)
        cos_angle = (dx0*dx1+dy0*dy1+dz0*dz1) / (r0*r1)
        # 弧长项
        arc_r = np.arccos(cos_angle)
        return arc_r

    def q_dihedral(self, p0, p1, p2, p3):
        """
        求解二面角
        """
        x0, y0, z0 = p0
        x1, y1, z1 = p1
        x2, y2, z2 = p2
        x3, y3, z3 = p3
        dx0 = x1 - x0
        dy0 = y1 - y0
        dz0 = z1 - z0
        dx1 = x2 - x1
        dy1 = y2 - y1
        dz1 = z2 - z1
        dx2 = x3 - x2
        dy2 = y3 - y2
        dz2 = z3 - z2
        # 平面法向量
        vx1 = dy0*dz1 - dz0*dy1
        vy1 = dz0*dx1 - dx0*dz1
        vz1 = dx0*dy1 - dy0*dx1
        vx2 = dy1*dz2 - dz1*dy2
        vy2 = dz1*dx2 - dx1*dz2
        vz2 = dx1*dy2 - dy1*dx2
        r_v1 = np.sqrt(vx1**2 + vy1**2 + vz1**2)
        # 方向符号
        normal = (vx1*dx2 + vy1*dy2 + vz1*dz2)
        normal = normal / np.abs(normal)
        r_v1 = np.sqrt(vx1**2 + vy1**2 + vz1**2)
        r_v2 = np.sqrt(vx2**2 + vy2**2 + vz2**2)
        cos_angle = (vx1*vx2+vy1*vy2+vz1*vz2) / (r_v1*r_v2)
        # 弧长项
        arc_r = np.arccos(cos_angle) * normal
        return arc_r

    def dq_bond(self, p0, p1):
        """
        半径长一阶导
        """
        result = derivatives.dq_b(p0, p1)
        return result

    def dq_angle(self, p0, p1, p2):
        """
        角度弧一阶导
        """
        result = derivatives.dq_a(p0, p1, p2)
        return result

    def dq_dihedral(self, p0, p1, p2, p3):
        """
        二面角弧一阶导
        """
        result = derivatives.dq_d(p0, p1, p2, p3)
        return result

    def d2q_bond(self, p0, p1):
        """
        半径长二阶导
        """
        result = derivatives.d2q_b(p0, p1)
        return result

    def d2q_angle(self, p0, p1, p2):
        """
        角度弧二阶导
        """
        result = derivatives.d2q_a(p0, p1, p2)
        return result

    def d2q_dihedral(self, p0, p1, p2, p3):
        """
        二面角弧二阶导
        """
        result = derivatives.d2q_d(p0, p1, p2, p3)
        return result

    def svd_inv(self, array, thresh, hermitian=False):
        """
        svd分解
        """
        U, S, Vt = np.linalg.svd(array, hermitian=hermitian)
        keep = S > thresh
        S_inv = np.zeros_like(S)
        S_inv[keep] = 1 / S[keep]
        return Vt.T.dot(np.diag(S_inv)).dot(U.T)

    def B_inv(self, B):
        return B.T.dot(self.svd_inv(B.dot(B.T), thresh=self.svd_inv_thresh, hermitian=True))

    def Bt_inv(self, B):
        return self.svd_inv(B.dot(B.T), thresh=self.svd_inv_thresh, hermitian=True).dot(B)

    def G_mat(self, B):
        """
        计算G矩阵
        """
        return B.dot(B.T)

    def G_inv(self, B):
        """
        计算G矩阵的逆
        """
        # return np.linalg.pinv(B.dot(B.T))
        G = self.G_mat(B)
        return G.T.dot(self.svd_inv(G.dot(G.T), thresh=self.svd_inv_thresh))

    def get_P(self, B):
        P = B.dot(self.B_inv(B))
        # P = self.G_mat(B).dot(self.G_inv(B))
        return P


if __name__ == "__main__":
    cart = [0.18413002, 1.25352350, 0.32249616,
            0.54078444, 0.24471350, 0.32249616,
            0.54080286, 1.75792169, 1.19614766,
            0.54080286, 1.75792169, 0.55115535,
            -0.88586998, 1.25353669, 0.32249616]
    cart = np.array(cart).reshape(-1, 3)
    connection = [[2, 1], [3, 1], [4, 1], [5, 1]]
    connection = [[item-1 for item in data] for data in connection]
    coord = RedundantCoordinate()
    redun_coord, b_mat = coord.cart2arc(cart, connection)
    print(redun_coord)
    print(b_mat)

